"""Interfaces with OpenMM."""

from contextlib import nullcontext
from pathlib import Path
from typing import TYPE_CHECKING, TextIO

from openff.toolkit import Molecule
from openff.utilities.utilities import has_package, requires_package

from openff.interchange._annotations import PositiveFloat
from openff.interchange.exceptions import (
    NegativeMassError,
    PluginCompatibilityError,
)
from openff.interchange.interop.openmm._import._import import from_openmm
from openff.interchange.interop.openmm._positions import to_openmm_positions
from openff.interchange.interop.openmm._topology import to_openmm_topology
from openff.interchange.smirnoff._base import SMIRNOFFCollection

if has_package("openmm"):
    import openmm
    import openmm.app

if TYPE_CHECKING:
    from openff.interchange import Interchange

__all__ = [
    "from_openmm",
    "to_openmm",
    "to_openmm_positions",
    "to_openmm_topology",
]


@requires_package("openmm")
def to_openmm_system(
    interchange: "Interchange",
    combine_nonbonded_forces: bool = False,
    add_constrained_forces: bool = False,
    ewald_tolerance: float = 1e-4,
    hydrogen_mass: PositiveFloat = 1.007947,
) -> "openmm.System":
    """
    Convert an Interchange to an OpenmM System.

    Parameters
    ----------
    interchange
        An OpenFF Interchange object
    combine_nonbonded_forces
        If True, an attempt will be made to combine all non-bonded interactions into a single openmm.NonbondedForce.
        If False, non-bonded interactions will be split across multiple forces.
    add_constrained_forces
        If True, add valence forces that might be overridden by constraints, i.e. call `addBond` or `addAngle`
        on a bond or angle that is fully constrained.
    ewald_tolerance
        The value passed to `NonbondedForce.setEwaldErrorTolerance`
    hydrogen_mass
        The mass to use for hydrogen atoms if not present in the topology. If non-trivially different
        than the default value, mass will be transferred from neighboring heavy atoms.

    Returns
    -------
    system
        The corresponding OpenMM System object

    """
    from openff.interchange.interop.openmm._gbsa import _process_gbsa
    from openff.interchange.interop.openmm._nonbonded import _process_nonbonded_forces
    from openff.interchange.interop.openmm._valence import (
        _process_angle_forces,
        _process_bond_forces,
        _process_constraints,
        _process_improper_torsion_forces,
        _process_torsion_forces,
    )

    for collection in interchange.collections.values():
        if collection.is_plugin:
            assert isinstance(collection, SMIRNOFFCollection)

            try:
                collection.check_openmm_requirements(combine_nonbonded_forces)
            except AssertionError as error:
                raise PluginCompatibilityError(
                    f"Collection of type {type(collection)} failed a compatibility check.",
                ) from error

    system = openmm.System()

    if interchange.box is not None:
        box = interchange.box.m_as("nanometer")
        system.setDefaultPeriodicBoxVectors(*box)

    particle_map = _process_nonbonded_forces(
        interchange,
        system,
        combine_nonbonded_forces=combine_nonbonded_forces,
        ewald_tolerance=ewald_tolerance,
    )

    constrained_pairs = _process_constraints(interchange, system, particle_map)

    _process_torsion_forces(interchange, system, particle_map)
    _process_improper_torsion_forces(interchange, system, particle_map)
    _process_angle_forces(
        interchange,
        system,
        add_constrained_forces=add_constrained_forces,
        constrained_pairs=constrained_pairs,
        particle_map=particle_map,
    )
    _process_bond_forces(
        interchange,
        system,
        add_constrained_forces=add_constrained_forces,
        constrained_pairs=constrained_pairs,
        particle_map=particle_map,
    )

    _process_gbsa(
        interchange,
        system,
    )

    # TODO: Apply HMR before or after final plugin touch point?
    _apply_hmr(
        system,
        interchange,
        hydrogen_mass=hydrogen_mass,
    )

    for collection in interchange.collections.values():
        if collection.is_plugin:
            assert isinstance(collection, SMIRNOFFCollection)

            try:
                collection.modify_openmm_forces(
                    interchange,
                    system,
                    add_constrained_forces=add_constrained_forces,
                    constrained_pairs=constrained_pairs,
                    particle_map=particle_map,
                )
            except NotImplementedError:
                continue

    system.addForce(openmm.CMMotionRemover())

    return system


to_openmm = to_openmm_system


@requires_package("openmm")
def _to_pdb(
    file_path: Path | str | TextIO,
    topology: "openmm.app.Topology",
    positions,
):
    from openff.units import ensure_quantity

    # Deal with the possibility of `StringIO`
    manager: nullcontext[TextIO] | TextIO  # MyPy needs some help here
    if isinstance(file_path, str | Path):
        manager = open(file_path, "w")
    else:
        manager = nullcontext(file_path)

    with manager as outfile:
        openmm.app.PDBFile.writeFile(
            topology=topology,
            positions=ensure_quantity(positions, "openmm"),
            file=outfile,
        )


@requires_package("openmm")
def _apply_hmr(
    system: "openmm.System",
    interchange: "Interchange",
    hydrogen_mass: PositiveFloat,
):
    if abs(hydrogen_mass - 1.008) < 1e-3:
        return

    water = Molecule.from_smiles("O")

    def _is_water(molecule: Molecule) -> bool:
        return molecule.is_isomorphic_with(water)

    _hydrogen_mass = hydrogen_mass * openmm.unit.dalton

    for bond in interchange.topology.bonds:
        heavy_atom, hydrogen_atom = bond.atoms

        if heavy_atom.atomic_number == 1:
            heavy_atom, hydrogen_atom = hydrogen_atom, heavy_atom

        # TODO: This should only skip rigid waters, even though HMR or flexible water is questionable
        if (
            (hydrogen_atom.atomic_number == 1)
            and (heavy_atom.atomic_number != 1)
            and not (_is_water(hydrogen_atom.molecule))
        ):
            hydrogen_index = interchange.topology.atom_index(hydrogen_atom)
            heavy_index = interchange.topology.atom_index(heavy_atom)
            heavy_mass = system.getParticleMass(heavy_index)

            # This will need to be wired up through the OpenFF-OpenMM particle index map
            # when virtual sites + HMR are supported
            mass_to_transfer = _hydrogen_mass - system.getParticleMass(hydrogen_index)

            if mass_to_transfer >= heavy_mass:
                raise NegativeMassError(
                    f"Particle with index {heavy_index} would have a negative mass after hydrogen "
                    "mass repartitioning. Consider transferring a smaller mass than "
                    f"{hydrogen_mass=}.",
                )

            system.setParticleMass(
                hydrogen_index,
                hydrogen_mass * openmm.unit.dalton,
            )

            system.setParticleMass(
                heavy_index,
                heavy_mass - mass_to_transfer,
            )



"""Custom Pydantic models."""

import abc
from typing import Any, Literal, cast

from pydantic import Field

from openff.interchange.pydantic import _BaseModel


class TopologyKey(_BaseModel, abc.ABC):
    """
    A unique identifier of a segment of a chemical topology.

    These refer to a single portion of a chemical graph, i.e. a single valence term,
    (a bond, angle, or dihedral) or a single atom. These target only the information in
    the chemical graph and do not store physics parameters. For example, a TopologyKey
    corresponding to a bond would store the indices of the two atoms that compose the
    bond, but not the force constant or equilibrium bond length as determined by the
    force field.

    Topology keys compare equal to (and hash the same as) tuples of their atom
    indices as long as their other fields are `None`.

    Examples
    --------
    Create a ``TopologyKey`` identifying some specific angle

    .. code-block:: pycon

        >>> from openff.interchange.models import TopologyKey
        >>> this_angle = TopologyKey(atom_indices=(2, 1, 3))
        >>> this_angle
        TopologyKey with atom indices (2, 1, 3)

    Create a ``TopologyKey`` indentifying just one atom

    .. code-block:: pycon

        >>> this_atom = TopologyKey(atom_indices=(4,))
        >>> this_atom
        TopologyKey with atom indices (4,)

    Compare a ``TopologyKey`` to a tuple containing the atom indices

    .. code-block:: pycon

        >>> key = TopologyKey(atom_indices=(0, 1))
        >>> key == (0, 1)
        True

    Index into a dictionary with a tuple

    .. code-block:: pycon

        >>> d = {TopologyKey(atom_indices=(0, 1)): "some_bond"}
        >>> d[0, 1]
        'some_bond'

    """

    atom_indices: tuple[int, ...] = Field(
        description="The indices of the atoms occupied by this interaction",
    )

    def _tuple(self) -> tuple:
        """Tuple representation of this key, which is overriden by most child classes."""
        return tuple(self.atom_indices)

    def __hash__(self) -> int:
        return hash(self._tuple())

    def __eq__(self, other: object) -> bool:
        if isinstance(other, tuple):
            return self._tuple() == other
        elif isinstance(other, TopologyKey):
            return self._tuple() == other._tuple()
        else:
            return NotImplemented

    def __repr__(self) -> str:
        return f"{self.__class__.__name__} with atom indices {self.atom_indices}"


class BondKey(TopologyKey):
    """
    A unique identifier of the atoms associated in a bond potential.

    Examples
    --------
    Index into a dictionary with a tuple

    .. code-block:: pycon

        >>> d = {
        ...     BondKey(atom_indices=(0, 1)): "some_bond",
        ...     BondKey(atom_indices=(1, 2), bond_order=1.5): "some_other_bond",
        ... }
        >>> d[0, 1]
        'some_bond'
        >>> d[(1, 2), 1.5]
        'some_other_bond'

    """

    atom_indices: tuple[int, int] = Field(
        description="The indices of the atoms occupied by this interaction",
    )

    bond_order: float | None = Field(
        None,
        description=(
            "If this key represents as topology component subject to interpolation between "
            "multiple parameters(s), the bond order determining the coefficients of the wrapped "
            "potentials."
        ),
    )

    def _tuple(self) -> tuple[int, int] | tuple[tuple[int, int], float]:
        if self.bond_order is None:
            return cast(tuple[int, int], self.atom_indices)
        else:
            return (
                cast(
                    tuple[int, int],
                    self.atom_indices,
                ),
                float(self.bond_order),
            )

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__} with atom indices {self.atom_indices}"
            f"{'' if self.bond_order is None else ', bond order ' + str(self.bond_order)}"
        )


class AngleKey(TopologyKey):
    """
    A unique identifier of the atoms associated in an angle potential.

    Examples
    --------
    Index into a dictionary with a tuple

    .. code-block:: pycon

        >>> d = {AngleKey(atom_indices=(0, 1, 2)): "some_angle"}
        >>> d[0, 1, 2]
        'some_angle'

    """

    atom_indices: tuple[int, int, int] = Field(
        description="The indices of the atoms occupied by this interaction",
    )

    def _tuple(self) -> tuple[int, int, int]:
        return cast(tuple[int, int, int], self.atom_indices)


class ProperTorsionKey(TopologyKey):
    """
    A unique identifier of the atoms associated in a proper torsion potential.

    Examples
    --------
    Index into a dictionary with a tuple

    .. code-block:: pycon

        >>> d = {
        ...     ProperTorsionKey(atom_indices=(0, 1, 2, 3)): "torsion1",
        ...     ProperTorsionKey(atom_indices=(0, 1, 2, 3), mult=2): "torsion2",
        ...     ProperTorsionKey(atom_indices=(5, 6, 7, 8), mult=2, phase=0.78, bond_order=1.5): "torsion3",
        ... }
        >>> d[0, 1, 2, 3]
        'torsion1'
        >>> d[(0, 1, 2, 3), 2, None, None]
        'torsion2'
        >>> d[(5, 6, 7, 8), 2, 0.78, 1.5]
        'torsion3'

    """

    atom_indices: tuple[int, int, int, int] = Field(
        description="The indices of the atoms occupied by this interaction",
    )

    mult: int | None = Field(
        None,
        description="The index of this duplicate interaction",
    )

    phase: float | None = Field(
        None,
        description="If this key represents as topology component subject to interpolation between "
        "multiple parameters(s), the phase determining the coefficients of the wrapped "
        "potentials.",
    )

    bond_order: float | None = Field(
        None,
        description=(
            "If this key represents as topology component subject to interpolation between "
            "multiple parameters(s), the bond order determining the coefficients of the wrapped "
            "potentials."
        ),
    )

    def _tuple(
        self,
    ) -> (
        tuple[int, int, int, int]
        | tuple[
            tuple[int, int, int, int],
            int | None,
            float | None,
            float | None,
        ]
    ):
        if self.mult is None and self.phase is None and self.bond_order is None:
            return cast(tuple[int, int, int, int], self.atom_indices)
        else:
            return (
                cast(
                    tuple[int, int, int, int],
                    self.atom_indices,
                ),
                self.mult,
                self.phase,
                self.bond_order,
            )

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__} with atom indices {self.atom_indices}"
            f"{'' if self.mult is None else ', mult ' + str(self.mult)}"
            f"{'' if self.bond_order is None else ', bond order ' + str(self.bond_order)}"
        )


class ImproperTorsionKey(ProperTorsionKey):
    """
    A unique identifier of the atoms associated in an improper torsion potential.

    The central atom is the first atom in the `atom_indices` tuple, or accessible via `get_central_atom_index`. Note
    that SMIRNOFF parameters list the central atom second and molecular simulation engines may list atoms in yet
    another order.

    Examples
    --------
    Index into a dictionary with a tuple

    .. code-block:: pycon

        >>> d = {
        ...     ImproperTorsionKey(atom_indices=(0, 1, 2, 3)): "torsion1",
        ...     ImproperTorsionKey(atom_indices=(0, 1, 2, 3), mult=2): "torsion2",
        ...     ImproperTorsionKey(atom_indices=(5, 6, 7, 8), mult=2, phase=0.78, bond_order=1.5): "torsion3",
        ... }
        >>> d[0, 1, 2, 3]
        'torsion1'
        >>> d[(0, 1, 2, 3), 2, None, None]
        'torsion2'
        >>> d[(5, 6, 7, 8), 2, 0.78, 1.5]
        'torsion3'

    """

    def get_central_atom_index(self) -> int:
        """Get the index of the central atom of this improper torsion."""
        return self.atom_indices[0]


class LibraryChargeTopologyKey(_BaseModel):
    """
    A unique identifier of the atoms associated with a library charge.
    """

    # TODO: Store all atoms associated with this charge?
    # TODO: Is there an upper bound on the number of atoms that can be associated with a LibraryChargeType?
    # TODO: Eventually rename this for coherence with `TopologyKey`
    this_atom_index: int

    @property
    def atom_indices(self) -> tuple[int]:
        """Alias for `this_atom_index`."""
        return (self.this_atom_index,)

    def __hash__(self) -> int:
        return hash((self.this_atom_index,))

    def __eq__(self, other) -> bool:
        return super().__eq__(other) or other == self.this_atom_index


class SingleAtomChargeTopologyKey(LibraryChargeTopologyKey):
    """
    Shim class for storing the result of charge_from_molecules.
    """

    extras: dict = dict()  # noqa: RUF012


class ChargeModelTopologyKey(_BaseModel):
    """Subclass of `TopologyKey` for use with charge models only."""

    this_atom_index: int
    partial_charge_method: str

    @property
    def atom_indices(self) -> tuple[int]:
        """Alias for `this_atom_index`."""
        return (self.this_atom_index,)

    def __hash__(self) -> int:
        return hash((self.this_atom_index, self.partial_charge_method))


class ChargeIncrementTopologyKey(_BaseModel):
    """Subclass of `TopologyKey` for use with charge increments only."""

    # TODO: Eventually rename this for coherence with `TopologyKey`
    this_atom_index: int
    other_atom_indices: tuple[int, ...]

    @property
    def atom_indices(self) -> tuple[int]:
        """Alias for `this_atom_index`."""
        return (self.this_atom_index,)

    def __hash__(self) -> int:
        return hash((self.this_atom_index, self.other_atom_indices))


class BaseVirtualSiteKey(TopologyKey):
    # TODO: Overriding the attribute of a parent class is clumsy, but less grief than
    #       having this not inherit from `TopologyKey`. It might be useful to just have
    #       orientation_atom_indices point to the same thing.
    atom_indices: tuple[int] | None = None  # type: ignore[assignment]

    orientation_atom_indices: tuple[int, ...] = Field(
        description="The indices of the 'orientation atoms' which are used to define the position "
        "of this virtual site. The first atom is the 'parent atom' which defines which atom the "
        "virtual site is 'attached' to.",
    )
    type: str = Field(description="The type of this virtual site parameter.")
    name: str = Field(description="The name of this virtual site parameter.")

    def _tuple(self) -> tuple:
        return (
            self.orientation_atom_indices,
            self.name,
            self.type,
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__} with orientation atom indices {self.orientation_atom_indices}"


class ImportedVirtualSiteKey(BaseVirtualSiteKey):
    """
    A unique identifier of a virtual site in the scope of a chemical topology.

    Meant to be used with data imported from OpenMM or other engines.

    Use the engine-specific identifier, like `openmm.ThreeParticleAverageSite`, in the "type" field.
    """

    pass


class SMIRNOFFVirtualSiteKey(BaseVirtualSiteKey):
    """A unique identifier of a SMIRNOFF virtual site in the scope of a chemical topology."""

    match: Literal["once", "all_permutations"] = Field(
        description="The `match` attribute of the associated virtual site type",
    )

    def _tuple(self) -> tuple:
        return (*super()._tuple(), self.match)


VirtualSiteKey = SMIRNOFFVirtualSiteKey


class PotentialKey(_BaseModel):
    """
    A unique identifier of an instance of physical parameters as applied to a segment of a chemical topology.

    These refer to a single term in a force field as applied to a single segment of a chemical
    topology, i.e. a single atom or dihedral. For example, a PotentialKey corresponding to a
    bond would store the the force constant and the equilibrium bond length as determined by
    the force field. These keys to not have direct knowledge of where in a topology they have been
    applied.

    Examples
    --------
    Create a PotentialKey corresponding to the parameter with id `b55` in OpenFF "Parsley" 1.0.0

    .. code-block:: pycon

        >>> from openff.interchange.models import PotentialKey
        >>> from openff.toolkit import ForceField
        >>> parsley = ForceField("openff-1.0.0.offxml")
        >>> param = parsley["Bonds"].get_parameter({"id": "b55"})[0]
        >>> bond_55 = PotentialKey(id=param.smirks)
        >>> bond_55
        PotentialKey associated with handler 'None' with id '[#16X4,#16X3:1]-[#8X2:2]'

    Create a PotentialKey corresponding to the angle parameters in OPLS-AA defined
    between atom types opls_135, opls_135, and opls_140

    .. code-block:: pycon

        >>> oplsaa_angle = PotentialKey(id="opls_135-opls_135-opls_140")
        >>> oplsaa_angle
        PotentialKey associated with handler 'None' with id 'opls_135-opls_135-opls_140'

    """

    id: str = Field(
        ...,
        description="A unique identifier of this potential, i.e. a SMARTS pattern or an atom type",
    )
    mult: int | None = Field(
        None,
        description="The index of this duplicate interaction",
    )
    associated_handler: str | None = Field(
        None,
        description="The type of handler this potential key is associated with, "
        "i.e. 'Bonds', 'vdW', or 'LibraryCharges",
    )
    bond_order: float | None = Field(
        None,
        description="If this is a key to a WrappedPotential interpolating multiple parameter(s), "
        "the bond order determining the coefficients of the wrapped potentials.",
    )
    virtual_site_type: str | None = Field(
        None,
        description="The 'type' of virtual site (i.e. `BondCharge`) this parameter is associated with.",
    )
    cosmetic_attributes: dict[str, Any] = Field(
        dict(),
        description="A dictionary of cosmetic attributes associated with this potential key.",
    )

    def __hash__(self) -> int:
        return hash((self.id, self.mult, self.associated_handler, self.bond_order))

    def __repr__(self) -> str:
        return (
            f"PotentialKey associated with handler '{self.associated_handler}' with id '{self.id}'"
            f"{'' if self.mult is None else ', mult ' + str(self.mult)}"
            f"{'' if self.bond_order is None else ', bond order ' + str(self.bond_order)}"
        )



import random
from importlib import resources
from math import exp

import numpy
import pytest
from openff.toolkit import ForceField, Molecule, Quantity, Topology
from openff.toolkit.typing.engines.smirnoff import VirtualSiteHandler
from openff.units import ensure_quantity
from openff.utilities import (
    get_data_file_path,
    has_package,
    requires_package,
    skip_if_missing,
)

from openff.interchange import Interchange
from openff.interchange._tests import (
    MoleculeWithConformer,
    get_protein,
    get_test_file_path,
    needs_gmx,
)
from openff.interchange.components.nonbonded import BuckinghamvdWCollection
from openff.interchange.components.potentials import Potential
from openff.interchange.drivers import get_gromacs_energies, get_openmm_energies
from openff.interchange.exceptions import GMXMdrunError, UnsupportedExportError
from openff.interchange.interop.gromacs._import._import import (
    _read_box,
    _read_coordinates,
    from_files,
)
from openff.interchange.models import PotentialKey, TopologyKey

if has_package("openmm"):
    import openmm.app
    import openmm.unit


@needs_gmx
class _NeedsGROMACS:
    pass


def parse_residue_ids(file) -> list[str]:
    """Parse residue IDs, and only the residue IDs, from a GROMACS .gro file."""
    with open(file) as f:
        ids = [line[:5].strip() for line in f.readlines()[2:-1]]

    return ids


class TestToGro(_NeedsGROMACS):
    @pytest.mark.xfail(reason="Broken")
    def test_residue_names(self, sage):
        """Reproduce issue #642."""
        # This could maybe just test the behavior of _convert?
        ligand = Molecule.from_smiles("CCO")
        ligand.generate_conformers(n_conformers=1)

        topology = ligand.to_topology()
        topology.box_vectors = Quantity([4, 4, 4], "nanometer")

        for atom in ligand.atoms:
            atom.metadata["residue_name"] = "LIG"

        sage.create_interchange(topology).to_gro("should_have_residue_names.gro")

        for line in open("should_have_residue_names.gro").readlines()[2:-2]:
            assert line[5:10] == "LIG  "

    @skip_if_missing("openmm")
    def test_tip4p_dimer(self, tip4p, water_dimer):
        tip4p.create_interchange(water_dimer).to_gro("_dimer.gro")

        positions = openmm.app.GromacsGroFile(
            "_dimer.gro",
        ).getPositions(asNumpy=True)

        assert positions.shape == (8, 3)

        assert not numpy.allclose(positions[3], positions[7])


@skip_if_missing("mdtraj")
@skip_if_missing("openmm")
class TestGROMACSGROFile(_NeedsGROMACS):
    try:
        _INTERMOL_PATH = resources.files(
            "intermol.tests.gromacs.unit_tests",
        )
    except ModuleNotFoundError:
        _INTERMOL_PATH = None

    @skip_if_missing("intermol")
    def test_load_gro(self):
        file = self._INTERMOL_PATH / "angle10_vacuum/angle10_vacuum.gro"

        positions = _read_coordinates(file)
        box = _read_box(file)

        openmm_gro = openmm.app.GromacsGroFile(file)
        openmm_positions = ensure_quantity(
            openmm_gro.getPositions(),
            "openff",
        )
        openmm_box = ensure_quantity(
            openmm_gro.getPeriodicBoxVectors(),
            "openff",
        )

        assert numpy.allclose(positions, openmm_positions)
        assert numpy.allclose(box, openmm_box)

    @skip_if_missing("intermol")
    @pytest.mark.skip("don't run parmed tests")
    def test_load_gro_nonstandard_precision(self):
        pytest.importorskip("intermol")

        file = self._INTERMOL_PATH / "lj3_bulk/lj3_bulk.gro"

        coords = _read_coordinates(file).m_as("nanometer")

        # OpenMM seems to assume a precision of 3. Use InterMol instead here.
        from intermol.gromacs.grofile_parser import GromacsGroParser

        intermol_gro = GromacsGroParser(file)
        intermol_gro.read()

        # InterMol stores positions as NumPy arrays _of openmm.unit.Quantity_
        # objects, so need to carefully convert everything through unitless
        # floats; converting through Pint-like quantities might produce a
        # funky double-wrapped thing
        def converter(x):
            return x.value_in_unit(openmm.unit.nanometer)

        other_coords = numpy.frompyfunc(converter, 1, 1)(intermol_gro.positions).astype(
            float,
        )

        assert numpy.allclose(coords, other_coords)

        # This file happens to have 12 digits of preicion; what really matters is that
        # the convential precision of 3 was not used.
        n_decimals = len(str(coords[0, 0]).split(".")[1])
        assert n_decimals == 12

    def test_vaccum_unsupported(self, sage):
        molecule = MoleculeWithConformer.from_smiles("CCO")

        out = Interchange.from_smirnoff(force_field=sage, topology=[molecule])

        assert out.box is None

        with pytest.raises(UnsupportedExportError, match="2020"):
            out.to_gro("tmp.gro")

    @pytest.mark.slow
    @pytest.mark.parametrize("offset_residue_ids", [True, False])
    @skip_if_missing("openmm")
    def test_residue_info(self, offset_residue_ids):
        """Test that residue information is passed through to .gro files."""
        import mdtraj

        protein = get_protein("MainChain_HIE")

        if offset_residue_ids:
            offset = random.randint(10, 20)

            for atom in protein.atoms:
                atom.metadata["residue_number"] = str(
                    int(atom.metadata["residue_number"]) + offset,
                )

            # Need to manually update residues _and_ atoms
            # https://github.com/openforcefield/openff-toolkit/issues/1921
            for residue in protein.residues:
                residue.residue_number = str(int(residue.residue_number) + offset)

        ff14sb = ForceField("ff14sb_off_impropers_0.0.3.offxml")

        out = Interchange.from_smirnoff(
            force_field=ff14sb,
            topology=[protein],
        )

        out.box = [4, 4, 4]

        out.to_gro("tmp.gro")

        mdtraj_topology = mdtraj.load("tmp.gro").topology

        for found_residue, original_residue in zip(
            mdtraj_topology.residues,
            out.topology.hierarchy_iterator("residues"),
        ):
            assert found_residue.name == original_residue.residue_name
            found_index = next(iter(original_residue.atoms)).metadata["residue_number"]
            assert str(found_residue.resSeq) == found_index

    def test_fill_in_residue_ids(self, sage):
        """Ensure that, if inputs have no residue_number, they are generated on-the-fly."""
        topology = Topology.from_molecules(
            [MoleculeWithConformer.from_smiles(smi) for smi in ["CC", "O", "C"]],
        )
        topology.box_vectors = Quantity([4, 4, 4], "nanometer")

        sage.create_interchange(topology).to_gro("fill.gro")

        expected_residue_ids = 8 * ["1"] + 3 * ["2"] + 5 * ["3"]

        found_residue_ids = parse_residue_ids("fill.gro")

        for expected, found in zip(expected_residue_ids, found_residue_ids):
            assert expected == found

    def test_atom_index_gt_100_000(self, water, sage):
        """Ensure that atom indices are written correctly for large numbers."""
        water.add_hierarchy_scheme(
            ("residue_name", "residue_number"),
            "residues",
        )

        topology = Topology.from_molecules(2 * [water])

        topology.box_vectors = Quantity([4, 4, 4], "nanometer")

        # Can't just update atoms' metadata, neeed to create these scheme/element objects
        # and need to also modify the residue objects themselves
        for molecule_index, molecule in enumerate(topology.molecules):
            for atom in molecule.atoms:
                atom.metadata["residue_number"] = str(molecule_index + 123_456)

        for residue_index, residue in enumerate(topology.residues):
            residue.residue_number = str(residue_index + 123_456)

        interchange = sage.create_interchange(topology)

        interchange.to_gro("large.gro")

        expected_residue_ids = 3 * ["23456"] + 3 * ["23457"]

        found_residue_ids = parse_residue_ids("large.gro")

        for expected, found in zip(expected_residue_ids, found_residue_ids):
            assert expected == found

    @pytest.mark.slow
    def test_atom_names_pdb(self):
        topology = get_protein("MainChain_ALA_ALA").to_topology()
        topology.box_vectors = Quantity([4, 4, 4], "nanometer")

        ff14sb = ForceField("ff14sb_off_impropers_0.0.3.offxml")

        Interchange.from_smirnoff(ff14sb, topology).to_gro(
            "atom_names.gro",
        )

        pdb_object = openmm.app.PDBFile(
            get_data_file_path("proteins/MainChain_ALA_ALA.pdb", "openff.toolkit"),
        )
        pdb_atom_names = [atom.name for atom in pdb_object.topology.atoms()]

        openmm_atom_names = openmm.app.GromacsGroFile("atom_names.gro").atomNames

        assert openmm_atom_names == pdb_atom_names


class TestGROMACS(_NeedsGROMACS):
    @pytest.mark.parametrize(
        "monolithic",
        [
            True,
            False,
        ],
    )
    @pytest.mark.parametrize(
        "smiles",
        [
            "C",
            "O=C=O",  # Adds unconstrained bonds without torsion(s)
            "CC",  # Adds a proper torsion term(s)
            # "C=O",  # Simplest molecule with any improper torsion
            "OC=O",  # Simplest molecule with a multi-term torsion
            # "CCOC",  # This hits t86, which has a non-1.0 idivf
            # "C1COC(=O)O1",  # This adds an improper, i2
        ],
    )
    def test_simple_roundtrip(
        self,
        sage_unconstrained,
        smiles,
        monolithic,
        monkeypatch,
    ):
        # Skip if using RDKit conformer, which does weird stuff around 180 deg
        pytest.importorskip("openeye.oechem")

        monkeypatch.setenv("INTERCHANGE_EXPERIMENTAL", "1")

        molecule = MoleculeWithConformer.from_smiles(smiles)
        molecule.name = molecule.to_hill_formula()
        topology = molecule.to_topology()

        out = sage_unconstrained.create_interchange(topology)

        out.box = [4, 4, 4]
        out.positions = numpy.round(molecule.conformers[0], 2)

        out.to_gromacs("out", decimal=3, monolithic=monolithic)

        converted = Interchange.from_gromacs("out.top", "out.gro")

        assert numpy.allclose(out.positions, converted.positions, atol=1e-3)
        assert numpy.allclose(out.box, converted.box)

        get_gromacs_energies(out).compare(
            get_gromacs_energies(converted),
            tolerances={
                "Bond": Quantity(0.002 * molecule.n_bonds, "kilojoule / mol"),
                "Electrostatics": Quantity(0.05, "kilojoule / mol"),
            },
        )

    def test_num_impropers(self, sage):
        parmed = pytest.importorskip("parmed")

        out = Interchange.from_smirnoff(
            sage,
            MoleculeWithConformer.from_smiles("CC1=CC=CC=C1").to_topology(),
        )

        out.box = Quantity(4 * numpy.eye(3), "nanometer")
        out.to_top("tmp.top")

        # Sanity check; toluene should have some improper(s)
        assert len(out["ImproperTorsions"].key_map) > 0

        struct = parmed.load_file("tmp.top")
        n_impropers_parmed = len([d for d in struct.dihedrals if d.improper])
        assert n_impropers_parmed == len(out["ImproperTorsions"].key_map)

    @pytest.mark.slow
    @skip_if_missing("intermol")
    @pytest.mark.skip(reason="Re-implement when SMIRNOFF supports more mixing rules")
    def test_set_mixing_rule(self, ethanol_top, sage):
        from intermol.gromacs.gromacs_parser import GromacsParser

        interchange = Interchange.from_smirnoff(force_field=sage, topology=ethanol_top)
        interchange.positions = numpy.zeros((ethanol_top.n_atoms, 3))
        interchange.to_gro("tmp.gro")

        interchange.box = [4, 4, 4]
        interchange.to_top("lorentz.top")
        lorentz = GromacsParser("lorentz.top", "tmp.gro").read()
        assert lorentz.combination_rule == "Lorentz-Berthelot"

        interchange["vdW"].mixing_rule = "geometric"

        interchange.to_top("geometric.top")
        geometric = GromacsParser("geometric.top", "tmp.gro").read()
        assert geometric.combination_rule == "Multiply-Sigeps"

    @pytest.mark.skip(reason="Re-implement when SMIRNOFF supports more mixing rules")
    def test_unsupported_mixing_rule(self, ethanol_top, sage):
        # TODO: Update this test when the model supports more mixing rules than GROMACS does
        interchange = Interchange.from_smirnoff(force_field=sage, topology=ethanol_top)
        interchange["vdW"].mixing_rule = "kong"

        with pytest.raises(UnsupportedExportError, match="rule `geometric` not compat"):
            interchange.to_top("out.top")

    @pytest.mark.slow
    @skip_if_missing("openmm")
    def test_residue_info(self, sage):
        """Test that residue information is passed through to .top files."""
        from openff.units.openmm import from_openmm

        parmed = pytest.importorskip("parmed")

        protein = get_protein("MainChain_HIE")

        box_vectors = from_openmm(
            openmm.app.PDBFile(
                get_data_file_path(
                    "proteins/MainChain_HIE.pdb",
                    "openff.toolkit",
                ),
            ).topology.getPeriodicBoxVectors(),
        )

        ff14sb = ForceField("ff14sb_off_impropers_0.0.3.offxml")

        out = Interchange.from_smirnoff(
            force_field=ff14sb,
            topology=[protein],
            box=box_vectors,
        )

        out.to_top("tmp.top")

        parmed_structure = parmed.load_file("tmp.top")

        for found_residue, original_residue in zip(
            parmed_structure.residues,
            out.topology.hierarchy_iterator("residues"),
        ):
            assert found_residue.name == original_residue.residue_name
            assert str(found_residue.number + 1) == original_residue.residue_number

    @pytest.mark.slow
    @pytest.mark.skip(
        reason="Update when energy reports intentionally support non-vdW handlers",
    )
    def test_argon_buck(self):
        """Test that Buckingham potentials are supported and can be exported"""
        from openff.interchange.smirnoff import SMIRNOFFElectrostaticsCollection

        mol = MoleculeWithConformer.from_smiles("[#18]", name="Argon")

        top = Topology.from_molecules([mol, mol])

        # http://www.sklogwiki.org/SklogWiki/index.php/Argon#Buckingham_potential
        A = Quantity(1.69e-8, "erg / mol * avogadro_number")
        B = Quantity(1 / 0.273, "1/angstrom")
        C = Quantity(102e-12, "erg/mol * avogadro_number * angstrom**6")

        r = Quantity(0.3, "nanometer")

        buck = BuckinghamvdWCollection()
        coul = SMIRNOFFElectrostaticsCollection(method="pme")

        pot_key = PotentialKey(id="[#18]")
        pot = Potential(parameters={"A": A, "B": B, "C": C})

        for atom in top.atoms:
            top_key = TopologyKey(atom_indices=(top.atom_index(atom),))
            buck.key_map.update({top_key: pot_key})

            coul.key_map.update({top_key: pot_key})
            coul.potentials.update(
                {pot_key: Potential(parameters={"charge": Quantity(0, "elementary_charge")})},
            )

        for molecule in top.molecules:
            molecule.partial_charges = Quantity(
                molecule.n_atoms * [0],
                "elementary_charge",
            )

        buck.potentials[pot_key] = pot

        out = Interchange()
        out.collections["Buckingham-6"] = buck
        out.collections["Electrostatics"] = coul
        out.topology = top
        out.box = Quantity([10, 10, 10], "nanometer")
        out.positions = Quantity([[0, 0, 0], [0.3, 0, 0]], "nanometer")
        out.to_gro("out.gro")
        out.to_top("out.top")

        omm_energies = get_openmm_energies(out, combine_nonbonded_forces=True)
        by_hand = A * exp(-B * r) - C * r**-6

        resid = omm_energies.energies["vdW"] - by_hand
        assert resid < Quantity(1e-5, "kilojoule / mol")

        # TODO: Add back comparison to GROMACS energies once GROMACS 2020+
        # supports Buckingham potentials
        with pytest.raises(GMXMdrunError):
            get_gromacs_energies(out, mdp="cutoff_buck")

    @pytest.mark.skip("Broken, unclear if cases like these are worth supporting")
    def test_nonconsecutive_isomorphic_molecules(self, sage_unconstrained):
        molecules = [Molecule.from_smiles(smiles) for smiles in ["CC", "CCO", "CC"]]

        for index, molecule in enumerate(molecules):
            molecule.generate_conformers(n_conformers=1)
            molecule.conformers[0] += Quantity(3 * [5 * index], "angstrom")

        topology = Topology.from_molecules(molecules)
        topology.box_vectors = Quantity([4, 4, 4], "nanometer")

        out = Interchange.from_smirnoff(sage_unconstrained, topology)

        get_gromacs_energies(out).compare(
            get_openmm_energies(out, combine_nonbonded_forces=True),
            {"Nonbonded": Quantity(0.5, "kilojoule / mole")},
        )

    @pytest.mark.parametrize("name", ["MOL0", "MOL999", ""])
    def test_exisiting_mol0_names_overwritten(self, name, sage, ethanol, cyclohexane):
        parmed = pytest.importorskip("parmed")

        ethanol.name = name
        cyclohexane.name = name

        sage.create_interchange(
            Topology.from_molecules([ethanol, cyclohexane]),
        ).to_top("tmp.top")

        assert [*parmed.load_file("tmp.top").molecules.keys()] == ["MOL0", "MOL1"]

    @pytest.mark.xfail(reason="very flaky in CI, not flaky locally")
    @pytest.mark.filterwarnings("ignore:Setting positions to None")
    @pytest.mark.parametrize("name", ["MOL0", "MOL222", ""])
    def test_roundtrip_with_combine(
        self,
        name,
        sage_unconstrained,
        ethanol,
        cyclohexane,
        monkeypatch,
    ):
        monkeypatch.setenv("INTERCHANGE_EXPERIMENTAL", "1")

        for molecule in [ethanol, cyclohexane]:
            molecule.name = name
            molecule.generate_conformers(n_conformers=1)

        # translate one by 1 nm
        cyclohexane._conformers = [
            cyclohexane.conformers[0] + Quantity([0, 0, 1], "nanometer"),
        ]

        ethanol_interchange = sage_unconstrained.create_interchange(
            ethanol.to_topology(),
        )
        ethanol_interchange.box = [4, 4, 4]

        cyclohexane_interchange = sage_unconstrained.create_interchange(
            cyclohexane.to_topology(),
        )
        cyclohexane_interchange.box = [4, 4, 4]

        combined = ethanol_interchange.combine(cyclohexane_interchange)

        ethanol_interchange.to_gromacs("ethanol", decimal=8)
        cyclohexane_interchange.to_gromacs("cyclohexane", decimal=8)
        combined.to_gromacs("combined", decimal=8)

        converted_ethanol = Interchange.from_gromacs(
            "ethanol.top",
            "ethanol.gro",
        )
        converted_cyclohexane = Interchange.from_gromacs(
            "cyclohexane.top",
            "cyclohexane.gro",
        )

        converted_combined = Interchange.from_gromacs(
            "combined.top",
            "combined.gro",
        )

        combined_from_converted = converted_ethanol.combine(converted_cyclohexane)
        combined_from_converted["vdW"].cutoff = combined["vdW"].cutoff
        converted_combined["vdW"].cutoff = combined["vdW"].cutoff

        assert numpy.allclose(converted_combined.positions, combined.positions)
        assert numpy.allclose(
            converted_combined.positions,
            combined_from_converted.positions,
        )
        assert numpy.allclose(converted_combined.box, combined.box)
        assert numpy.allclose(converted_combined.box, combined_from_converted.box)

        get_gromacs_energies(combined).compare(
            get_gromacs_energies(converted_combined),
            tolerances={
                "Electrostatics": Quantity(0.05, "kilojoule / mole"),
            },
        )
        get_gromacs_energies(combined).compare(
            get_gromacs_energies(combined_from_converted),
            tolerances={
                "Electrostatics": Quantity(0.05, "kilojoule / mole"),
            },
        )


class TestGROMACSMetadata(_NeedsGROMACS):
    @skip_if_missing("openmm")
    @skip_if_missing("mdtraj")
    @pytest.mark.slow
    def test_atom_names_pdb(self):
        topology = get_protein("MainChain_ALA_ALA").to_topology()
        topology.box_vectors = Quantity([4, 4, 4], "nanometer")

        ff14sb = ForceField("ff14sb_off_impropers_0.0.3.offxml")

        ff14sb.create_interchange(topology).to_gromacs("atom_names")

        pdb_object = openmm.app.PDBFile(
            get_data_file_path(
                "proteins/MainChain_ALA_ALA.pdb",
                "openff.toolkit",
            ),
        )
        openmm_object = openmm.app.GromacsTopFile("atom_names.top")

        pdb_atom_names = [atom.name for atom in pdb_object.topology.atoms()]

        openmm_atom_names = [atom.name for atom in openmm_object.topology.atoms()]

        assert openmm_atom_names == pdb_atom_names


class TestSettles(_NeedsGROMACS):
    def test_settles_units(self, monkeypatch, water):
        """Reproduce issue #720."""
        monkeypatch.setenv("INTERCHANGE_EXPERIMENTAL", "1")

        water.name = "WAT"

        interchange = ForceField("openff-2.1.0.offxml").create_interchange(
            water.to_topology(),
        )
        interchange.box = [10, 10, 10]
        interchange.to_gromacs(
            prefix="settles",
        )

        system = from_files("settles.top", "settles.gro")

        settles = system.molecule_types["WAT"].settles[0]

        assert settles.oxygen_hydrogen_distance.m_as("angstrom") == pytest.approx(
            0.9572,
        )
        assert settles.hydrogen_hydrogen_distance.m_as("angstrom") == pytest.approx(
            1.513900654525,
        )


@pytest.mark.slow
@requires_package("openmm")
class TestCommonBoxes(_NeedsGROMACS):
    @pytest.mark.parametrize(
        "pdb_file",
        [
            get_test_file_path("cube.pdb").as_posix(),
            get_test_file_path("dodecahedron.pdb").as_posix(),
            get_test_file_path("octahedron.pdb").as_posix(),
        ],
    )
    def test_common_boxes(self, pdb_file):
        original_box_vectors = openmm.app.PDBFile(
            pdb_file,
        ).topology.getPeriodicBoxVectors()
        # TODO: Regenerate test files to be simpler and smaller, no need to use a protein
        force_field = ForceField(
            "openff-2.1.0.offxml",
            "ff14sb_off_impropers_0.0.3.offxml",
        )

        topology = Topology.from_pdb(pdb_file)

        force_field.create_interchange(topology).to_gro(pdb_file.replace("pdb", "gro"))

        parsed_box_vectors = openmm.app.GromacsGroFile(
            pdb_file.replace("pdb", "gro"),
        ).getPeriodicBoxVectors()

        numpy.testing.assert_allclose(
            numpy.array(original_box_vectors.value_in_unit(openmm.unit.nanometer)),
            numpy.array(parsed_box_vectors.value_in_unit(openmm.unit.nanometer)),
        )


class TestMergeAtomTypes(_NeedsGROMACS):
    @pytest.mark.parametrize(
        "smiles",
        [
            "CC",  # Two identical carbons
            "C1CCCCC1",  # Identical carbons in the ring
            "C1[C@@H](N)C[C@@H](N)CC1",  # Identical carbons and nitrogens in the ring
        ],
    )
    def test_energies_with_merging_atom_types(self, sage, smiles):
        """
        Tests #962
        """
        molecule = MoleculeWithConformer.from_smiles(smiles)
        molecule.name = molecule.to_hill_formula()
        topology = molecule.to_topology()

        out = Interchange.from_smirnoff(force_field=sage, topology=topology)
        out.box = [4, 4, 4]
        out.positions = molecule.conformers[0]

        get_gromacs_energies(out).compare(
            get_gromacs_energies(out, _merge_atom_types=True),
            tolerances={
                "Bond": Quantity(0.002 * molecule.n_bonds, "kilojoule / mol"),
                "Electrostatics": Quantity(0.05, "kilojoule / mol"),
            },
        )

    @pytest.mark.slow
    @pytest.mark.parametrize(
        "smiles",
        [
            "CC",  # Two identical carbons
            "C1CCCCC1",  # Identical carbons in the ring
            "C1[C@@H](N)C[C@@H](N)CC1",  # Identical carbons and nitrogens in the ring
        ],
    )
    def test_simple_roundtrip_with_merging_atom_types(
        self,
        sage_unconstrained,
        smiles,
        monkeypatch,
    ):
        """
        Tests #962
        """
        monkeypatch.setenv("INTERCHANGE_EXPERIMENTAL", "1")

        molecule = MoleculeWithConformer.from_smiles(smiles)
        molecule.name = molecule.to_hill_formula()
        topology = molecule.to_topology()

        out = sage_unconstrained.create_interchange(topology)
        out.box = [4, 4, 4]
        out.positions = molecule.conformers[0]

        out.to_top("out.top")
        out.to_top("out_merged.top", _merge_atom_types=True)
        out.to_gro("out.gro", decimal=3)

        converted = Interchange.from_gromacs("out.top", "out.gro")
        converted_merged = Interchange.from_gromacs(
            "out_merged.top",
            "out.gro",
        )

        assert numpy.allclose(converted.positions, converted_merged.positions)
        assert numpy.allclose(converted.box, converted_merged.box)

        get_gromacs_energies(converted_merged).compare(
            get_gromacs_energies(converted),
            tolerances={
                "Bond": Quantity(0.002 * molecule.n_bonds, "kilojoule / mol"),
                "Electrostatics": Quantity(0.05, "kilojoule / mol"),
            },
        )

    @pytest.mark.parametrize(
        "molecule_list",
        [
            ["CC", "CCO"],
            ["CC", "CCCC"],
            ["c1ccccc1", "CC", "CCO", "O"],
            ["c1ccncc1", "n1cnccc1", "[nH]1cccc1"],
        ],
    )
    def test_merge_atom_types_of_similar_molecules(
        self,
        molecule_list,
        sage,
    ):
        parmed = pytest.importorskip("parmed")

        topology = Topology.from_molecules(
            [Molecule.from_smiles(smi) for smi in molecule_list],
        )

        out = sage.create_interchange(topology)

        for merge in [True, False]:
            out.to_top(f"{merge}.top", _merge_atom_types=merge)

        not_merged = parmed.load_file("False.top")
        merged = parmed.load_file("True.top")

        # These are (n_atoms x 1) lists of parameters, which should be
        # read from [ atoms ] section and cross-referenced to [ atomtypes ]
        for attr in ["sigma", "epsilon", "charge"]:
            assert [getattr(atom, attr) for atom in not_merged.atoms] == [getattr(atom, attr) for atom in merged.atoms]


class TestGROMACSVirtualSites(_NeedsGROMACS):
    @pytest.fixture
    def sigma_hole_type(self, sage):
        """A handler with a bond charge virtual site on a C-Cl bond."""
        return VirtualSiteHandler.VirtualSiteBondChargeType(
            name="EP",
            smirks="[#6:1]-[#17:2]",
            distance=Quantity(1.4 * "angstrom"),
            type="BondCharge",
            match="once",
            charge_increment1=Quantity(0.1, "elementary_charge"),
            charge_increment2=Quantity(0.2, "elementary_charge"),
        )

    @pytest.fixture
    def sage_with_monovalent_lone_pair(self, sage):
        """Fixture that loads an SMIRNOFF XML for argon"""
        virtual_site_handler = VirtualSiteHandler(version=0.3)

        carbonyl_type = VirtualSiteHandler.VirtualSiteType(
            name="EP",
            smirks="[O:1]=[C:2]-[*:3]",
            distance=Quantity(0.3, "angstrom"),
            type="MonovalentLonePair",
            match="all_permutations",
            outOfPlaneAngle=Quantity(0.0, "degree"),
            inPlaneAngle=Quantity(120.0, "degree"),
            charge_increment1=Quantity(0.05, "elementary_charge"),
            charge_increment2=Quantity(0.1, "elementary_charge"),
            charge_increment3=Quantity(0.15, "elementary_charge"),
        )

        virtual_site_handler.add_parameter(parameter=carbonyl_type)
        sage.register_parameter_handler(virtual_site_handler)

        return sage

    @pytest.mark.xfail
    @skip_if_missing("parmed")
    def test_sigma_hole_example(self, sage_with_sigma_hole):
        """Test that a single-molecule sigma hole example runs"""
        parmed = pytest.importorskip("parmed")

        molecule = MoleculeWithConformer.from_smiles("CCl", name="Chloromethane")

        out = Interchange.from_smirnoff(
            force_field=sage_with_sigma_hole,
            topology=molecule.to_topology(),
        )
        out.box = [4, 4, 4]
        out.positions = molecule.conformers[0]

        # TODO: Sanity-check reported energies
        get_gromacs_energies(out)

        out.to_top("sigma.top")
        gmx_top = parmed.load_file("sigma.top")

        assert abs(numpy.sum([p.charge for p in gmx_top.atoms])) < 1e-3

    @pytest.mark.slow
    def test_carbonyl_example(self, sage_with_planar_monovalent_carbonyl, ethanol):
        """Test that a single-molecule planar carbonyl example can run 0 steps."""
        ethanol.generate_conformers(n_conformers=1)

        hexanal = MoleculeWithConformer.from_smiles("CCCCCC=O")
        hexanal._conformers[0] += Quantity("2 nanometer")

        topology = Topology.from_molecules([ethanol, hexanal])
        topology.box_vectors = Quantity([4, 4, 4], "nanometer")

        get_gromacs_energies(
            sage_with_planar_monovalent_carbonyl.create_interchange(topology),
        )

    @skip_if_missing("openmm")
    def test_tip4p_charge_neutrality(self, tip4p, water_dimer):
        tip4p.create_interchange(water_dimer).to_top("_dimer.top")

        system = openmm.app.GromacsTopFile("_dimer.top").createSystem()

        assert system.getForce(0).getNumParticles() == 8

        charges = [system.getForce(0).getParticleParameters(i)[0]._value for i in range(4)]

        assert sum(charges) == pytest.approx(0.0)



import pytest
from openff.toolkit import Quantity
from openff.utilities import skip_if_missing

from openff.interchange._tests import MoleculeWithConformer, requires_ambertools
from openff.interchange.constants import kcal_mol
from openff.interchange.drivers import get_amber_energies, get_openmm_energies


@requires_ambertools
@skip_if_missing("openmm")
@pytest.mark.parametrize(
    "smiles",
    ["c1ccccc1", "C1=CC=C2C(=C1)C=CC3=C2C=CC4=CC=CC=C43"],
)
def test_polycyclic_nonbonded(smiles, sage_no_switch):
    molecule = MoleculeWithConformer.from_smiles(smiles)
    topology = molecule.to_topology()
    topology.box_vectors = Quantity([10, 10, 10], "nanometer")

    interchange = sage_no_switch.create_interchange(topology)

    openmm_vdw = get_openmm_energies(
        interchange,
        combine_nonbonded_forces=False,
    ).energies["vdW"]

    amber_vdw = get_amber_energies(interchange).energies["vdW"]

    assert openmm_vdw.m == pytest.approx(
        amber_vdw.m,
        rel=1e-4,
    ), f"{openmm_vdw.m_as(kcal_mol)=}, {amber_vdw.m_as(kcal_mol)=}"



"""Unit tests for `InterchangeTopology`."""

import pytest

from openff.interchange.common._topology import InterchangeTopology
from openff.interchange.exceptions import NoPositionsError


@pytest.mark.parametrize("conversion_method", ["json", "dict"])
def test_from_toolkit_topology(alanine_dipeptide, conversion_method):
    match conversion_method:
        case "json":
            topology = InterchangeTopology.from_json(
                alanine_dipeptide.to_json(),
            )
        case "dict":
            topology = InterchangeTopology.from_dict(
                alanine_dipeptide.to_dict(),
            )

    assert topology.n_atoms == alanine_dipeptide.n_atoms
    assert topology.n_bonds == alanine_dipeptide.n_bonds
    assert topology.n_molecules == alanine_dipeptide.n_molecules

    for atom_a, atom_b in zip(topology.atoms, alanine_dipeptide.atoms):
        assert atom_a.atomic_number == atom_b.atomic_number

    for exception in [NoPositionsError, NotImplementedError]:
        with pytest.raises(exception):
            topology.get_positions()

        with pytest.raises(exception):
            topology.clear_positions()

        with pytest.raises(exception):
            topology.set_positions(alanine_dipeptide.molecule(0).conformers[0])



import packaging.version
import pytest

from openff.interchange.interop.lammps.export.provenance import get_lammps_version


def test_get_lammps_version():
    pytest.importorskip("lammps")

    assert get_lammps_version() > packaging.version.Version("2020")

    assert get_lammps_version() < packaging.version.Version("2100")



"""
A wrapper around PACKMOL. Adapted from OpenFF Evaluator v0.4.3.
"""

from openff.packmol import (
    RHOMBIC_DODECAHEDRON,
    RHOMBIC_DODECAHEDRON_XYHEX,
    UNIT_CUBE,
    pack_box,
    solvate_topology,
    solvate_topology_nonwater,
)
from openff.packmol._packmol import (
    _box_from_density,
    _box_vectors_are_in_reduced_form,
    _build_input_file,
    _center_topology_at,
    _check_add_positive_mass,
    _check_box_shape_shape,
    _compute_brick_from_box_vectors,
    _create_molecule_pdbs,
    _create_solute_pdb,
    _find_packmol,
    _get_packmol_version,
    _iter_lattice_vecs,
    _load_positions,
    _max_dist_between_points,
    _range_neg_pos,
    _scale_box,
    _unit_vec,
    _validate_inputs,
    _wrap_into,
    _wrap_into_brick,
)

__all__ = (
    "RHOMBIC_DODECAHEDRON",
    "RHOMBIC_DODECAHEDRON_XYHEX",
    "UNIT_CUBE",
    "_box_from_density",
    "_box_vectors_are_in_reduced_form",
    "_build_input_file",
    "_center_topology_at",
    "_check_add_positive_mass",
    "_check_box_shape_shape",
    "_compute_brick_from_box_vectors",
    "_create_molecule_pdbs",
    "_create_solute_pdb",
    "_find_packmol",
    "_get_packmol_version",
    "_iter_lattice_vecs",
    "_load_positions",
    "_max_dist_between_points",
    "_range_neg_pos",
    "_scale_box",
    "_unit_vec",
    "_validate_inputs",
    "_wrap_into",
    "_wrap_into_brick",
    "pack_box",
    "solvate_topology",
    "solvate_topology_nonwater",
)



import numpy
from numpy.typing import NDArray
from openff.toolkit import Quantity, Topology

from openff.interchange import Interchange
from openff.interchange.common._positions import _infer_positions
from openff.interchange.components.potentials import Collection
from openff.interchange.foyer._guard import has_foyer
from openff.interchange.foyer._nonbonded import (
    FoyerElectrostaticsHandler,
    FoyerVDWHandler,
)
from openff.interchange.foyer._valence import (
    FoyerHarmonicAngleHandler,
    FoyerHarmonicBondHandler,
    FoyerPeriodicImproperHandler,
    FoyerPeriodicProperHandler,
    FoyerRBImproperHandler,
    FoyerRBProperHandler,
)
from openff.interchange.models import TopologyKey

if has_foyer:
    try:
        from foyer import Forcefield
    except ModuleNotFoundError:
        pass

_CollectionAlias = type[Collection]


def get_handlers_callable() -> dict[str, _CollectionAlias]:
    """Map Foyer-style handlers from string identifiers."""
    return {
        "vdW": FoyerVDWHandler,
        "Electrostatics": FoyerElectrostaticsHandler,
        "Bonds": FoyerHarmonicBondHandler,
        "Angles": FoyerHarmonicAngleHandler,
        "RBTorsions": FoyerRBProperHandler,
        "RBImpropers": FoyerRBImproperHandler,
        "ProperTorsions": FoyerPeriodicProperHandler,
        "ImproperTorsions": FoyerPeriodicImproperHandler,
    }


def _create_interchange(
    force_field: "Forcefield",
    topology: Topology,
    box: Quantity | None = None,
    positions: Quantity | None = None,
) -> Interchange:
    interchange = Interchange(topology=Topology(topology))

    interchange.positions = _infer_positions(interchange.topology, positions)

    interchange.box = interchange.topology.box_vectors if box is None else box

    # This block is from a mega merge, unclear if it's still needed
    for name, handler_class in get_handlers_callable().items():
        interchange.collections[name] = handler_class()  # type: ignore[call-arg]

    vdw_handler = interchange["vdW"]
    vdw_handler.scale_14 = force_field.lj14scale
    vdw_handler.store_matches(force_field, topology=interchange.topology)  # type: ignore[attr-defined]
    vdw_handler.store_potentials(force_field=force_field)  # type: ignore[attr-defined]

    atom_slots = vdw_handler.key_map

    electrostatics = interchange["Electrostatics"]
    electrostatics.scale_14 = force_field.coulomb14scale
    electrostatics.store_charges(  # type: ignore[attr-defined]
        atom_slots=atom_slots,
        force_field=force_field,
    )

    for name, handler in interchange.collections.items():
        if name not in ["vdW", "Electrostatics"]:
            handler.store_matches(atom_slots, topology=interchange.topology)  # type: ignore[attr-defined]
            handler.store_potentials(force_field)  # type: ignore[attr-defined]

    # TODO: Populate .mdconfig, but only after a reasonable number of state mutations have been tested

    charges = electrostatics.charges

    for molecule in interchange.topology.molecules:
        molecule_charges: NDArray[numpy.int_] = numpy.asarray(
            [charges[TopologyKey(atom_indices=(interchange.topology.atom_index(atom),))].m for atom in molecule.atoms],
        )
        molecule.partial_charges = Quantity(molecule_charges, "elementary_charge")

    interchange.collections["vdW"] = vdw_handler
    interchange.collections["Electrostatics"] = electrostatics

    return interchange



import pytest
from openff.packmol import solvate_topology_nonwater
from openff.packmol._packmol import _find_packmol
from openff.toolkit import Quantity

from openff.interchange._tests import MoleculeWithConformer
from openff.interchange.drivers import get_openmm_energies


@pytest.mark.slow
@pytest.mark.skipif(_find_packmol() is None, reason="PACKMOL not found")
def test_solvate_ligand_in_nonwater(sage):
    """
    Test that a ligand can be solvated in a non-water solvent with a sane
    starting point, like setting up a SFE run.

    See uses of `solvate_topology` somewhere in openfe_skunkworks
    """
    pytest.importorskip("openmm")

    # ibuprofen in hexanol
    ligand = MoleculeWithConformer.from_smiles(
        "CC(C)CC1=CC=C(C=C1)C(C)C(=O)O",
        allow_undefined_stereo=True,
    )
    solvent = MoleculeWithConformer.from_smiles("CCCCCCO")

    topology = solvate_topology_nonwater(
        topology=ligand.to_topology(),
        solvent=solvent,
        padding=Quantity(0.8, "nanometer"),
        target_density=Quantity(800, "kilogram / meter ** 3"),
    )

    interchange = sage.create_interchange(topology)

    packed_energy = get_openmm_energies(interchange)

    interchange.minimize(force_tolerance=Quantity(100, "kJ / mol / nm"))

    assert get_openmm_energies(interchange).total_energy < packed_energy.total_energy



import numpy
import pytest
from openff.toolkit import ForceField, Molecule, Quantity, Topology
from openff.utilities import has_package, skip_if_missing

from openff.interchange import Interchange
from openff.interchange._tests import (
    HAS_GROMACS,
    HAS_LAMMPS,
    MoleculeWithConformer,
    get_test_file_path,
    needs_gmx,
    needs_lmp,
)
from openff.interchange.constants import kj_mol
from openff.interchange.drivers import get_openmm_energies
from openff.interchange.foyer._guard import has_foyer

if has_package("openmm"):
    import openmm
    import openmm.unit

if HAS_GROMACS:
    from openff.interchange.drivers.gromacs import (
        _get_mdp_file,
        _process,
        _run_gmx_energy,
        get_gromacs_energies,
    )
if HAS_LAMMPS:
    from openff.interchange.drivers.lammps import get_lammps_energies


@pytest.mark.skipif(not has_foyer, reason="Foyer is not installed")
class TestEnergies:
    @pytest.fixture(scope="session")
    def oplsaa(self):
        import foyer

        return foyer.forcefields.load_OPLSAA()

    @skip_if_missing("mbuild")
    @needs_gmx
    @needs_lmp
    @pytest.mark.xfail
    @pytest.mark.parametrize("constrained", [True, False])
    @pytest.mark.parametrize("mol_smi", ["C"])  # ["C", "CC"]
    def test_energies_single_mol(self, constrained, sage, sage_unconstrained, mol_smi):
        import mbuild

        molecule = MoleculeWithConformer.from_smiles(mol_smi, name="FOO")

        force_field = sage if constrained else sage_unconstrained

        interchange = Interchange.from_smirnoff(force_field, [molecule])

        interchange.collections["Electrostatics"].periodic_potential = "cutoff"

        molecule.to_file("out.xyz", file_format="xyz")
        compound = mbuild.load("out.xyz")
        packed_box = mbuild.fill_box(
            compound=compound,
            n_compounds=1,
            box=mbuild.Box(lengths=[10, 10, 10]),
        )

        positions = Quantity(packed_box.xyz, "nanometer")
        interchange.positions = positions

        omm_energies = get_openmm_energies(interchange, round_positions=8)

        mdp = "cutoff_hbonds" if constrained else "auto"
        # Compare GROMACS writer and OpenMM export
        gmx_energies = get_gromacs_energies(interchange, mdp=mdp)

        tolerances = {
            "Electrostatics": 2 * openmm.unit.kilojoule_per_mole,
            "vdW": 2 * openmm.unit.kilojoule_per_mole,
            "Nonbonded": 2 * openmm.unit.kilojoule_per_mole,
        }

        gmx_energies.compare(
            omm_energies,
            tolerances,
        )

        if not constrained:
            other_energies = get_openmm_energies(
                interchange,
                round_positions=8,
                hard_cutoff=True,
                electrostatics=True,
            )
            lmp_energies = get_lammps_energies(interchange)
            tolerances = {
                "vdW": 5.0 * openmm.unit.kilojoule_per_mole,
                "Electrostatics": 5.0 * openmm.unit.kilojoule_per_mole,
            }
            lmp_energies.compare(other_energies, tolerances)

    @needs_gmx
    @pytest.mark.skipif(not has_foyer, reason="Foyer is not installed")
    @skip_if_missing("mbuild")
    def test_process_rb_torsions(self, oplsaa):
        """Test that the GROMACS driver reports Ryckaert-Bellemans torsions"""
        from mbuild import Box

        from openff.interchange.components.mbuild import offmol_to_compound

        ethanol = MoleculeWithConformer.from_smiles("CCO")
        ethanol.generate_unique_atom_names()

        my_compound = offmol_to_compound(ethanol)
        my_compound.box = Box(lengths=[4, 4, 4])

        struct = oplsaa.apply(my_compound)

        struct.save("eth.top", overwrite=True)
        struct.save("eth.gro", overwrite=True)

        # Get single-point energies using GROMACS
        oplsaa_energies = _process(
            _run_gmx_energy(
                top_file="eth.top",
                gro_file="eth.gro",
                mdp_file=_get_mdp_file("default"),
            ),
        )

        assert oplsaa_energies.energies["RBTorsion"].m != 0.0

    @needs_gmx
    def test_gmx_14_energies_exist(self, sage):
        # TODO: Make sure 1-4 energies are accurate, not just existent

        # Use a molecule with only one 1-4 interaction, and
        # make it between heavy atoms because H-H 1-4 are weak
        mol = MoleculeWithConformer.from_smiles("ClC#CCl", name="HPER")

        out = Interchange.from_smirnoff(sage, [mol])
        out.positions = mol.conformers[0]
        out.box = 3 * [10]

        # Put this molecule in a large box with cut-off electrostatics
        # to prevent it from interacting with images of itself
        out["Electrostatics"].periodic_potential = "cutoff"

        gmx_energies = get_gromacs_energies(out)

        # The only possible non-bonded interactions should be from 1-4 intramolecular interactions
        assert gmx_energies.energies["vdW"].m != 0.0
        assert gmx_energies.energies["Electrostatics"].m != 0.0

        # TODO: It would be best to save the 1-4 interactions, split off into vdW and Electrostatics
        # in the energies. This might be tricky/intractable to do for engines that are not GROMACS

    @needs_gmx
    @needs_lmp
    def test_cutoff_electrostatics(self):
        ion_ff = ForceField(get_test_file_path("ions.offxml"))
        ions = Topology.from_molecules(
            [
                Molecule.from_smiles("[#3+]"),
                Molecule.from_smiles("[#17-]"),
            ],
        )
        out = Interchange.from_smirnoff(ion_ff, ions)
        out.box = Quantity([4, 4, 4], "nanometer")

        gmx = []
        lmp = []

        for d in numpy.linspace(0.75, 0.95, 5):
            positions = Quantity(numpy.zeros((2, 3)), "nanometer")
            positions[1, 0] = Quantity(d, "nanometer")
            out.positions = positions

            out["Electrostatics"].periodic_potential = "cutoff"
            gmx.append(
                get_gromacs_energies(out, mdp="auto").energies["Electrostatics"].m,
            )
            lmp.append(
                get_lammps_energies(out).energies["Electrostatics"].m_as("kilojoule / mol"),
            )

        assert numpy.sum(numpy.sqrt(numpy.square(numpy.asarray(lmp) - numpy.asarray(gmx)))) < 1e-3

    @pytest.mark.parametrize(
        "smi",
        [
            "c1cc(ccc1c2ccncc2)O",
            "c1cc(ccc1c2ccncc2)[O-]",
            "c1cc(c(cc1O)Cl)c2cc[nH+]cc2",
        ],
    )
    @needs_gmx
    @pytest.mark.slow
    def test_interpolated_parameters(self, smi):
        xml_ff_bo_all_heavy_bonds = """<?xml version='1.0' encoding='ASCII'?>
        <SMIRNOFF version="0.3" aromaticity_model="OEAroModel_MDL">
          <Bonds version="0.3" fractional_bondorder_method="AM1-Wiberg" fractional_bondorder_interpolation="linear">
            <Bond smirks="[!#1:1]~[!#1:2]" id="bbo1"
                k_bondorder1="100.0 * kilocalories_per_mole/angstrom**2"
                k_bondorder2="1000.0 * kilocalories_per_mole/angstrom**2"
                length_bondorder1="1.5 * angstrom"
                length_bondorder2="1.0 * angstrom"/>
          </Bonds>
        </SMIRNOFF>
        """

        molecule = MoleculeWithConformer.from_smiles(smi)

        forcefield = ForceField(
            "openff-2.0.0.offxml",
            xml_ff_bo_all_heavy_bonds,
        )

        out = Interchange.from_smirnoff(forcefield, [molecule])
        out.box = Quantity([4, 4, 4], "nanometer")
        out.positions = molecule.conformers[0]

        for key in ["Bond", "Torsion"]:
            interchange_energy = get_openmm_energies(
                out,
                combine_nonbonded_forces=True,
            ).energies[key]

            gromacs_energy = get_gromacs_energies(out).energies[key]
            energy_diff = abs(interchange_energy - gromacs_energy).m_as(kj_mol)

            if energy_diff < 1e-6:
                pass
            elif energy_diff < 1e-2:
                pytest.xfail(
                    f"Found {key} energy difference of {energy_diff} kJ/mol between GROMACS and OpenMM exports",
                )
            else:
                pytest.fail(
                    f"Found {key} energy difference of {energy_diff} kJ/mol between GROMACS and OpenMM exports",
                )

    @needs_gmx
    def test_rb_torsion_energies(self, monkeypatch):
        monkeypatch.setenv("INTERCHANGE_EXPERIMENTAL", "1")

        interchange = Interchange.from_gromacs(
            get_test_file_path("gromacs/rb_torsions.top"),
            get_test_file_path("gromacs/rb_torsions.gro"),
        )

        openmm_energy = get_openmm_energies(interchange)["RBTorsion"]
        gromacs_energy = get_gromacs_energies(interchange)["RBTorsion"]

        assert abs(openmm_energy - gromacs_energy).m_as("kilojoule_per_mole") < 1e-4



import abc
from collections.abc import Iterable
from typing import Any, Literal

from openff.toolkit import Quantity
from pydantic import Field, PrivateAttr, computed_field

from openff.interchange._annotations import _DistanceQuantity, _ElementaryChargeQuantity
from openff.interchange.components.potentials import Collection
from openff.interchange.models import (
    LibraryChargeTopologyKey,
    SingleAtomChargeTopologyKey,
    TopologyKey,
    VirtualSiteKey,
)


class _NonbondedCollection(Collection, abc.ABC):
    type: str = "nonbonded"

    cutoff: _DistanceQuantity = Field(
        Quantity(10.0, "angstrom"),
        description="The distance at which pairwise interactions are truncated",
    )

    scale_12: float = Field(
        0.0,
        description="The scaling factor applied to 1-2 interactions",
    )
    scale_13: float = Field(
        0.0,
        description="The scaling factor applied to 1-3 interactions",
    )
    scale_14: float = Field(
        0.5,
        description="The scaling factor applied to 1-4 interactions",
    )
    scale_15: float = Field(
        1.0,
        description="The scaling factor applied to 1-5 interactions",
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @classmethod
    def __init_subclass__(cls, **kwargs):
        """Hack to get electrostatics subclasses to have private attributes."""
        if "Electrostatics" in cls.__name__:
            cls._charges = dict()
            cls._charges_cached = False

        return super().__pydantic_init_subclass__(**kwargs)


class vdWCollection(_NonbondedCollection):
    """Handler storing vdW potentials."""

    type: Literal["vdW"] = "vdW"

    expression: Literal["4*epsilon*((sigma/r)**12-(sigma/r)**6)"] = "4*epsilon*((sigma/r)**12-(sigma/r)**6)"

    mixing_rule: str = Field(
        "lorentz-berthelot",
        description="The mixing rule (combination rule) used in computing pairwise vdW interactions",
    )

    switch_width: _DistanceQuantity = Field(
        Quantity(1.0, "angstrom"),
        description="The width over which the switching function is applied",
    )

    periodic_method: Literal["cutoff", "no-cutoff", "pme"] = Field("cutoff")
    nonperiodic_method: Literal["cutoff", "no-cutoff", "pme"] = Field("no-cutoff")

    @classmethod
    def default_parameter_values(cls) -> Iterable[float]:
        """Per-particle parameter values passed to Force.addParticle()."""
        return 1.0, 0.0

    @classmethod
    def potential_parameters(cls) -> Iterable[str]:
        """Return a list of names of parameters included in each potential in this colletion."""
        return "sigma", "epsilon"


class ElectrostaticsCollection(_NonbondedCollection):
    """Handler storing electrostatics interactions."""

    type: Literal["Electrostatics"] = "Electrostatics"

    expression: Literal["coul"] = "coul"

    periodic_potential: Literal[
        "Ewald3D-ConductingBoundary",
        "cutoff",
        "no-cutoff",
        "reaction-field",
    ] = Field("Ewald3D-ConductingBoundary")
    nonperiodic_potential: Literal[
        "Coulomb",
        "cutoff",
        "no-cutoff",
        "reaction-field",
    ] = Field("Coulomb")
    exception_potential: Literal["Coulomb"] = Field("Coulomb")

    _charges: dict[Any, _ElementaryChargeQuantity] = PrivateAttr()
    _charges_cached: bool = PrivateAttr(default=False)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def charges(
        self,
    ) -> dict[TopologyKey | LibraryChargeTopologyKey | VirtualSiteKey, _ElementaryChargeQuantity]:
        """Get the total partial charge on each atom, including virtual sites."""
        if len(self._charges) == 0 or self._charges_cached is False:
            self._charges.update(self._get_charges(include_virtual_sites=False))
            self._charges_cached = True

        return self._charges

    def get_charge_array(self, include_virtual_sites: bool = False) -> Quantity:
        """
        Return a one-dimensional array-like of atomic charges, ordered topologically.

        If virtual sites are present in the system, `NotImplementedError` is raised.
        """
        if include_virtual_sites:
            raise NotImplementedError("Not yet implemented with virtual sites")

        if VirtualSiteKey in {type(key) for key in self.key_map}:
            raise NotImplementedError(
                "Not yet implemented when virtual sites are present, even with `include_virtual_sites=False`.",
            )

        return Quantity.from_list([q for _, q in sorted(self.charges.items(), key=lambda x: x[0].atom_indices)])

    def _get_charges(
        self,
        include_virtual_sites: bool = False,
    ) -> dict[TopologyKey | VirtualSiteKey | LibraryChargeTopologyKey, _ElementaryChargeQuantity]:
        if include_virtual_sites:
            raise NotImplementedError()

        return {
            topology_key: self.potentials[potential_key].parameters["charge"]
            for topology_key, potential_key in self.key_map.items()
        }


def _simpler_charges(
    collection: ElectrostaticsCollection,
) -> dict[TopologyKey, Quantity]:
    """Unify SingleAtomChargeTopologyKey to TopologyKey."""
    return_dict = dict()

    for key, val in collection.charges.items():
        if isinstance(key, SingleAtomChargeTopologyKey):
            new_key = TopologyKey(atom_indices=key.atom_indices)
            return_dict[new_key] = val
        elif isinstance(key, TopologyKey):
            return_dict[key] = val
        else:
            raise KeyError(f"Unexpected key of type {type(key)=}")

    return return_dict



import os
from functools import wraps
from inspect import cleandoc

from openff.interchange.exceptions import ExperimentalFeatureException


def experimental(func):
    """
    Decorate a function as experimental, requiring environment variable opt-in.

    To use the wrapped function, set the environment variable INTERCHANGE_EXPERIMENTAL=1.

    """
    firstline, _, remainder = cleandoc(func.__doc__).partition("\n")

    func.__doc__ = (
        "🧪 "
        f"{firstline}"
        "\n\n.. admonition:: Experimental\n\n    This object is experimental and should not be used in production.\n"
        f"{remainder}"
    )

    @wraps(func)
    def wrapper(*args, **kwargs):
        if os.environ.get("INTERCHANGE_EXPERIMENTAL", "0") != "1":
            raise ExperimentalFeatureException(
                f"\n\tFunction or method {func.__name__} is experimental. This feature is not "
                "complete, not yet reliable, and/or needs more testing to be considered suitable "
                "for production.\n"
                "\tTo use this feature on a provisional basis, set the environment variable "
                "INTERCHANGE_EXPERIMENTAL=1.",
            )

        return func(*args, **kwargs)

    return wrapper



import pytest
from openff.toolkit import Quantity

from openff.interchange._tests import needs_gmx
from openff.interchange.drivers.gromacs import get_gromacs_energies
from openff.interchange.interop.gromacs._import._import import from_files
from openff.interchange.interop.gromacs._interchange import to_interchange


class TestToInterchange:
    @needs_gmx
    def test_torsion_multiplicities_stored_in_keys(
        self,
        ethanol,
        sage_unconstrained,
        monkeypatch,
    ):
        """Reproduce issue #978."""
        parmed = pytest.importorskip("parmed")

        monkeypatch.setenv("INTERCHANGE_EXPERIMENTAL", "1")

        ethanol.generate_conformers(n_conformers=1)
        topology = ethanol.to_topology()
        topology.box_vectors = Quantity([4, 4, 4], "nanometer")

        out = sage_unconstrained.create_interchange(topology)
        out.to_gromacs("ethanol", decimal=8)

        parsed = from_files("ethanol.top", "ethanol.gro")

        num_torsions_in_file = len(parmed.load_file("ethanol.top").dihedrals)

        # Check that dihedrals are not override in the intermediate GROMACSSystem object
        assert num_torsions_in_file == len(parsed.molecule_types["MOL0"].dihedrals)

        # Check that GROMACSSystem -> Interchange doesn't lose information
        assert num_torsions_in_file == len(
            to_interchange(parsed)["ProperTorsions"].key_map,
        )

        get_gromacs_energies(
            out,
        ).compare(
            get_gromacs_energies(to_interchange(parsed)),
            tolerances={"Electrostatics": Quantity(0.01, "kilojoule / mole")},
        )



import warnings
from collections import defaultdict
from typing import TYPE_CHECKING, Union

from openff.toolkit import Quantity, Topology
from openff.units import ensure_quantity
from openff.units.openmm import from_openmm as from_openmm_
from openff.utilities.utilities import has_package, requires_package
from pydantic import ValidationError

from openff.interchange.common._nonbonded import ElectrostaticsCollection, vdWCollection
from openff.interchange.common._topology import InterchangeTopology
from openff.interchange.common._valence import (
    AngleCollection,
    BondCollection,
    ConstraintCollection,
    ProperTorsionCollection,
)
from openff.interchange.exceptions import NoPositionsError, UnsupportedImportError
from openff.interchange.interop.openmm._import._virtual_sites import _convert_virtual_sites
from openff.interchange.interop.openmm._import.compat import _check_compatible_inputs
from openff.interchange.models import ImportedVirtualSiteKey
from openff.interchange.warnings import MissingPositionsWarning

if has_package("openmm"):
    import openmm
    import openmm.app
    import openmm.unit

if TYPE_CHECKING:
    import openmm
    import openmm.app
    import openmm.unit

    from openff.interchange import Interchange


@requires_package("openmm")
def from_openmm(
    *,
    system: "openmm.System",
    topology: Union["openmm.app.Topology", Topology],
    positions: Quantity | None = None,
    box_vectors: Quantity | None = None,
) -> "Interchange":
    """Create an Interchange object from OpenMM data."""
    from openff.interchange import Interchange

    _check_compatible_inputs(system=system, topology=topology)

    if isinstance(topology, openmm.app.Topology):
        from openff.interchange.components.toolkit import _simple_topology_from_openmm

        openff_topology = _simple_topology_from_openmm(topology, system)

        if topology.getPeriodicBoxVectors() is not None:
            openff_topology.box_vectors = from_openmm_(topology.getPeriodicBoxVectors())

        # OpenMM topologies do not store positions

        particle_map = openff_topology._particle_map

    elif isinstance(topology, (Topology, InterchangeTopology)):
        # could split this out into separate cases, but behavior is similar
        openff_topology = topology

        particle_map = {index: index for index in range(topology.n_atoms)}
        try:
            positions = openff_topology.get_positions()
        except NoPositionsError:
            # fall back to positions argument,
            # InterchangeTopology.get_positions() raises this
            pass

    else:
        raise ValueError(
            f"Could not process `topology` argument of type {type(topology)=}.",
        )

    interchange = Interchange(topology=openff_topology)

    try:
        interchange.topology._molecule_virtual_site_map = openff_topology._molecule_virtual_site_map
    except AttributeError:
        interchange.topology._molecule_virtual_site_map = defaultdict(list)

    # TODO: Actually build up the VirtualSiteCollection, maybe using _molecule_virtual_site_map

    constraints = _convert_constraints(system, particle_map)

    if constraints is not None:
        interchange.collections["Constraints"] = constraints

    virtual_sites = _convert_virtual_sites(system, openff_topology, particle_map)

    if virtual_sites is not None:
        interchange.collections["VirtualSites"] = virtual_sites

    for force in system.getForces():
        if isinstance(force, openmm.NonbondedForce):
            vdw, coul = _convert_nonbonded_force(force, particle_map)
            interchange.collections["vdW"] = vdw
            interchange.collections["Electrostatics"] = coul
        elif isinstance(force, openmm.HarmonicBondForce):
            bonds = _convert_harmonic_bond_force(force)
            interchange.collections["Bonds"] = bonds
        elif isinstance(force, openmm.HarmonicAngleForce):
            angles = _convert_harmonic_angle_force(force)
            interchange.collections["Angles"] = angles
        elif isinstance(force, openmm.PeriodicTorsionForce):
            proper_torsions = _convert_periodic_torsion_force(force)
            interchange.collections["ProperTorsions"] = proper_torsions
        elif isinstance(force, openmm.CMMotionRemover):
            pass
        else:
            raise UnsupportedImportError(
                f"Unsupported OpenMM Force type ({type(force)}) found.",
            )

    if positions is None:
        warnings.warn(
            "Nothing was passed to the `positions` argument, are you sure you don't want to pass positions?",
            MissingPositionsWarning,
        )

    else:
        assert len(positions) == len(particle_map)

        interchange.positions = ensure_quantity(positions, "openff")[  # type: ignore[index]
            [key for key, val in particle_map.items() if isinstance(val, int)]
        ]

    if box_vectors is not None:
        _box_vectors = box_vectors

    elif interchange.topology.box_vectors is not None:
        _box_vectors = interchange.topology.box_vectors

    else:
        # If there is no box argument passed and the topology is non-periodic
        # and the system does not have default box vectors, it'll end up as None
        _box_vectors = from_openmm_(system.getDefaultPeriodicBoxVectors())

    # TODO: Does this run through the Interchange.box validator?
    interchange.box = _box_vectors

    try:
        num_physics_bonds = len(interchange["Bonds"].key_map)
    except LookupError:
        num_physics_bonds = 0

    if interchange.topology.n_bonds > num_physics_bonds:
        # There are probably missing (physics) bonds from rigid waters. The topological
        # bonds are probably processed correctly.
        _fill_in_rigid_water_bonds(interchange)

    return interchange


def _convert_constraints(
    system: "openmm.System",
    particle_map: dict[int, int | ImportedVirtualSiteKey],
) -> ConstraintCollection | None:
    from openff.interchange.components.potentials import Potential
    from openff.interchange.models import BondKey, PotentialKey

    if system.getNumConstraints() == 0:
        return None

    constraints = ConstraintCollection()

    # Map the unique distances (float, implicitly nanometer) to indices used for deduplication
    unique_distances: dict[float, int] = {
        distance: index
        for index, distance in enumerate(
            {
                system.getConstraintParameters(index)[2].value_in_unit(
                    openmm.unit.nanometer,
                )
                for index in range(system.getNumConstraints())
            },
        )
    }

    _keys: dict[float, PotentialKey] = dict()

    for distance, index in unique_distances.items():
        potential_key = PotentialKey(id=f"Constraint{index}")
        _keys[distance] = potential_key
        constraints.potentials[potential_key] = Potential(
            parameters={"distance": Quantity(distance, "nanometer")},
        )

    for index in range(system.getNumConstraints()):
        atom1, atom2, _distance = system.getConstraintParameters(index)

        distance = _distance.value_in_unit(openmm.unit.nanometer)

        constraints.key_map[BondKey(atom_indices=(particle_map[atom1], particle_map[atom2]))] = _keys[distance]

    return constraints


def _convert_nonbonded_force(
    force: "openmm.NonbondedForce",
    particle_map: dict[int, int | ImportedVirtualSiteKey],
) -> tuple[vdWCollection, ElectrostaticsCollection]:
    from openff.units.openmm import from_openmm as from_openmm_quantity

    from openff.interchange.components.potentials import Potential
    from openff.interchange.models import PotentialKey, TopologyKey

    if force.getNonbondedMethod() != 4:
        raise UnsupportedImportError(
            "Importing from OpenMM only currently supported with `openmm.NonbondedForce.PME`.",
        )

    vdw = vdWCollection()
    electrostatics = ElectrostaticsCollection(version=0.4, scale_14=0.833333)

    n_parametrized_particles = force.getNumParticles()

    for idx in range(n_parametrized_particles):
        charge, sigma, epsilon = force.getParticleParameters(idx)

        try:
            top_key = TopologyKey(atom_indices=(particle_map[idx],))
        except ValidationError:
            top_key: ImportedVirtualSiteKey = particle_map[idx]  # type: ignore[no-redef]

        pot = Potential(
            parameters={
                "sigma": from_openmm_quantity(sigma),
                "epsilon": from_openmm_quantity(epsilon),
            },
        )

        pot_key = PotentialKey(id=f"{idx}", associated_handler="vdW")
        vdw.key_map.update({top_key: pot_key})
        vdw.potentials.update({pot_key: pot})

        # This quacks like it's from a library charge, but tracks that it's
        # not actually coming from a source
        pot_key = PotentialKey(id=f"{idx}", associated_handler="ExternalSource")
        electrostatics.key_map.update({top_key: pot_key})
        electrostatics.potentials.update(
            {pot_key: Potential(parameters={"charge": from_openmm_quantity(charge)})},
        )

    if force.getNonbondedMethod() == 4:
        vdw.cutoff = force.getCutoffDistance()
        electrostatics.cutoff = force.getCutoffDistance()
    else:
        raise UnsupportedImportError(
            f"Parsing a non-bonded force of type {type(force)} with {force.getNonbondedMethod()} not yet supported.",
        )

    if force.getUseSwitchingFunction():
        vdw.switch_width = vdw.cutoff - from_openmm_quantity(
            force.getSwitchingDistance(),
        )
    else:
        vdw.switch_width = 0.0 * vdw.cutoff.units

    return vdw, electrostatics


def _convert_harmonic_bond_force(
    force: "openmm.HarmonicBondForce",
) -> BondCollection:
    from openff.units.openmm import from_openmm as from_openmm_quantity

    from openff.interchange.common._valence import BondCollection
    from openff.interchange.components.potentials import Potential
    from openff.interchange.models import BondKey, PotentialKey

    bonds = BondCollection()

    n_parametrized_bonds = force.getNumBonds()

    for idx in range(n_parametrized_bonds):
        atom1, atom2, length, k = force.getBondParameters(idx)
        top_key = BondKey(atom_indices=(atom1, atom2))
        pot_key = PotentialKey(id=f"{atom1}-{atom2}", associated_handler="Bonds")
        pot = Potential(
            parameters={
                "length": from_openmm_quantity(length),
                "k": from_openmm_quantity(k),
            },
        )

        bonds.key_map.update({top_key: pot_key})
        bonds.potentials.update({pot_key: pot})

    return bonds


def _convert_harmonic_angle_force(
    force: "openmm.HarmonicAngleForce",
) -> AngleCollection:
    from openff.units.openmm import from_openmm as from_openmm_quantity

    from openff.interchange.common._valence import AngleCollection
    from openff.interchange.components.potentials import Potential
    from openff.interchange.models import AngleKey, PotentialKey

    angles = AngleCollection()

    n_parametrized_angles = force.getNumAngles()

    for idx in range(n_parametrized_angles):
        atom1, atom2, atom3, angle, k = force.getAngleParameters(idx)
        top_key = AngleKey(atom_indices=(atom1, atom2, atom3))
        pot_key = PotentialKey(
            id=f"{atom1}-{atom2}-{atom3}",
            associated_handler="Angles",
        )
        pot = Potential(
            parameters={
                "angle": from_openmm_quantity(angle),
                "k": from_openmm_quantity(k),
            },
        )

        angles.key_map.update({top_key: pot_key})
        angles.potentials.update({pot_key: pot})

    return angles


def _convert_periodic_torsion_force(
    force: "openmm.PeriodicTorsionForce",
) -> ProperTorsionCollection:
    # TODO: Can impropers be separated out from a PeriodicTorsionForce?
    # Maybe by seeing if a quartet is in mol/top.propers or .impropers
    from openff.units.openmm import from_openmm as from_openmm_quantity

    from openff.interchange.common._valence import ProperTorsionCollection
    from openff.interchange.components.potentials import Potential
    from openff.interchange.models import PotentialKey, ProperTorsionKey

    proper_torsions = ProperTorsionCollection()

    n_parametrized_torsions = force.getNumTorsions()

    for idx in range(n_parametrized_torsions):
        atom1, atom2, atom3, atom4, per, phase, k = force.getTorsionParameters(idx)
        # TODO: Process layered torsions
        # TODO: Check if this torsion is an improper
        top_key = ProperTorsionKey(atom_indices=(atom1, atom2, atom3, atom4), mult=0)
        while top_key in proper_torsions.key_map:
            top_key.mult = top_key.mult + 1  # type: ignore[operator]

        pot_key = PotentialKey(
            id=f"{atom1}-{atom2}-{atom3}-{atom4}",
            mult=top_key.mult,
            associated_handler="ProperTorsions",
        )
        pot = Potential(
            parameters={
                "periodicity": Quantity(int(per), "dimensionless"),
                "phase": from_openmm_quantity(phase),
                "k": from_openmm_quantity(k),
                "idivf": Quantity(1, "dimensionless"),
            },
        )

        proper_torsions.key_map.update({top_key: pot_key})
        proper_torsions.potentials.update({pot_key: pot})

    return proper_torsions


def _fill_in_rigid_water_bonds(interchange: "Interchange"):
    from openff.toolkit.topology._mm_molecule import Molecule, _SimpleMolecule

    from openff.interchange.components.potentials import Potential
    from openff.interchange.models import AngleKey, BondKey, PotentialKey

    if "Bonds" not in interchange.collections:
        interchange.collections.update({"Bonds": BondCollection()})

    if "Angles" not in interchange.collections:
        interchange.collections.update({"Angles": AngleCollection()})

    simple_water = _SimpleMolecule.from_molecule(Molecule.from_smiles("O"))

    rigid_water_bond_key = PotentialKey(id="rigid_water", associated_handler="Bonds")
    rigid_water_bond = Potential(
        parameters={
            "length": Quantity("1.0 angstrom"),
            "k": Quantity("50,000.0 kcal/mol/angstrom**2"),
        },
    )

    rigid_water_angle_key = PotentialKey(id="rigid_water", associated_handler="Angles")
    rigid_water_angle = Potential(
        parameters={
            "angle": Quantity("104.5 degree"),
            "k": Quantity("1.0 kcal/mol/rad**2"),
        },
    )

    interchange["Bonds"].potentials.update(
        {PotentialKey(id="rigid_water", associated_handler="Bonds"): rigid_water_bond},
    )

    interchange["Angles"].potentials.update(
        {
            PotentialKey(
                id="rigid_water",
                associated_handler="Angles",
            ): rigid_water_angle,
        },
    )

    for molecule in interchange.topology.molecules:
        if not molecule.is_isomorphic_with(simple_water):
            continue

        for bond in molecule.bonds:
            bond_key = BondKey(
                atom_indices=(
                    interchange.topology.atom_index(bond.atom1),
                    interchange.topology.atom_index(bond.atom2),
                ),
            )

            if bond_key not in interchange["Bonds"].key_map:
                # add 1 A / 50,000 kcal/mol/A2 force constant
                interchange["Bonds"].key_map.update({bond_key: rigid_water_bond_key})

        for angle in molecule.angles:
            angle_key = AngleKey(
                atom_indices=(
                    interchange.topology.atom_index(angle[0]),
                    interchange.topology.atom_index(angle[1]),
                    interchange.topology.atom_index(angle[2]),
                ),
            )

            if angle_key not in interchange["Angles"].key_map:
                # add very flimsy force constant, since equilibrium angles differ
                # across models
                interchange["Angles"].key_map.update({angle_key: rigid_water_angle_key})



import pytest
from openff.toolkit import Quantity
from openff.utilities.testing import skip_if_missing

from openff.interchange._tests import MoleculeWithConformer

# Some of these tests are basically copied from [1] with an extra step into GROMACS files on disk.  Might be useful to
# collapse these in order to avoid divergence in the future.
# [1] unit_tests/interop/openmm/test_virtual_sites.py


@skip_if_missing("openmm")
class TestBondChargeVirtualSite:
    @pytest.mark.parametrize(
        (
            "distance_",
            "expected_w1",
            "expected_w2",
        ),
        [
            (0.8, 1.5, -0.5),
            (0.0, 1.0, 0.0),
            (-0.8, 0.5, 0.5),
            (-1.6, 0.0, 1.0),
            (-2.4, -0.5, 1.5),
        ],
    )
    def test_bond_charge_weights(
        self,
        sage_with_bond_charge,
        water,
        distance_,
        expected_w1,
        expected_w2,
    ):
        import openmm
        import openmm.app

        sage_with_bond_charge["VirtualSites"].get_parameter(
            parameter_attrs={
                "smirks": "[#6:2]-[#17X1:1]",
            },
        )[0].distance = Quantity(
            distance_,
            "angstrom",
        )

        sage_with_bond_charge.create_interchange(
            MoleculeWithConformer.from_mapped_smiles(
                "[H:3][C:2]([H:4])([H:5])[Cl:1]",
            ).to_topology(),
        ).to_top(f"_foo{distance_}.top")

        system = openmm.app.GromacsTopFile(f"_foo{distance_}.top").createSystem()

        assert system.getNumParticles() == 6

        for index in range(5):
            assert system.getParticleMass(index)._value > 0.0

            try:
                system.getVirtualSite(index)
            except openmm.OpenMMException as error:
                # This method raises an error, not sure if there's another method
                # that just returns a boolean as to whether or not it's a virtual site
                assert "This particle is not a virtual site" in str(error)

        assert system.getParticleMass(5)._value == 0.0

        virtual_site = system.getVirtualSite(5)

        assert isinstance(virtual_site, openmm.TwoParticleAverageSite)

        assert virtual_site.getParticle(0) == 0
        assert virtual_site.getParticle(1) == 1

        assert virtual_site.getWeight(0) == pytest.approx(expected_w1)
        assert virtual_site.getWeight(1) == pytest.approx(expected_w2)



"""Script used to convert molecules.smi to molecules.sdf."""

from openff.toolkit.topology.molecule import Molecule
from rdkit import Chem

mols = Molecule.from_file("molecules.smi", allow_undefined_stereo=True)

with Chem.SDWriter("molecules.sdf") as w:
    for mol in mols:
        try:
            mol.generate_conformers(n_conformers=1)
        except:  # noqa
            continue
        mol.assign_partial_charges(partial_charge_method="am1bcc")
        rdmol = mol.to_rdkit()
        w.write(rdmol)



import subprocess

import packaging.version
import polars

from openff.interchange import __version__

most_recent_release = packaging.version.Version(
    subprocess.check_output(
        ["git", "describe", "--abbrev=0", "--tags"],
    ).decode(),
)


try:
    current_version = f"Current version (commit {__version__.split('+g')[1].split('.')[0]})"
except IndexError:
    current_version = __version__

for header, version in zip(
    ("Most recent release", current_version),
    (most_recent_release, __version__),
):
    print(f"\n=== {header} ({version}) ===")
    systems = {
        "Mixed solvent": f"{version}/mixed_solvent.csv",
        "Ligand in water": f"{version}/ligand_in_water.csv",
        "Protein in water": f"{version}/protein_in_water.csv",
        "500-mer PEG": f"{version}/large_peg.csv",
    }

    for system_name, file_path in systems.items():
        data_frame = polars.read_csv(file_path)
        print(
            f"{system_name} benchmark summary:\n"
            f"\tOpenMM export time (s): "
            f"{data_frame['openmm_time'].mean():.3f} ± {data_frame['openmm_time'].std():.3f}\n"
            f"\tGROMACS export time (s): "
            f"{data_frame['gromacs_time'].mean():.3f} ± {data_frame['gromacs_time'].std():.3f}\n",
        )



"""Test charge assignment logging."""

import logging
import re
from collections import defaultdict

import pytest
from openff.toolkit import ForceField, Molecule, NAGLToolkitWrapper, RDKitToolkitWrapper, Topology
from openff.toolkit.utils import OPENEYE_AVAILABLE
from openff.toolkit.utils.toolkit_registry import ToolkitRegistry, toolkit_registry_manager

from openff.interchange._tests import get_protein

"""
Hierarchy is
1. Match molecules with preset charges
2. Match chemical environment against library charges
3. Run NAGLCharges (if present in the FF)
3. Run charge method in ChargeIncrementModel section (if present in the FF) and then
   match chemical environment against charge increments
4. Run AM1-BCC (or a variant) on remaining molecules (if present in the FF)

Test cases
----------

Note that each case uses pre-NAGL Sage

0. Sage with a ligand in vacuum
* Ligand gets AM1-BCC

1. Sage with a ligand in water/ions
* Water gets library charges
* Ions get library charges
* Non-water solvent gets AM1-BCC
* Ligands get AM1-BCC

2. Sage with a ligand in non-water solvent
* Non-water solvent gets AM1-BCC
* Ligands get AM1-BCC

2. Sage with a ligand in mixed solvent
* Water gets library charges
* Ions get library charges
* Non-water solvent gets AM1-BCC
* Ligands get AM1-BCC

4. ff14SB with Sage
* Protein gets library charges
* Water gets library charges
* Ions get library charges
* Non-water solvent gets AM1-BCC
* Ligands get AM1-BCC

5. ff14SB with Sage and preset charges on Protein A
* Protein A gets preset charges
* Other proteins get library charges
* Water gets library charges
* Ions get library charges
* Non-water solvent gets AM1-BCC
* Ligands get AM1-BCC

6. Sage with ligand and OPC water
* Water gets library charges
* Water virtual sites get ??
* Ions get library charges
* Non-water solvent gets AM1-BCC
* Ligands get AM1-BCC

7. Sage with preset charges on ligand A
* Water gets library charges
* Ions get library charges
* Non-water solvent gets AM1-BCC
* Ligand A gets preset charges
* Other ligands get AM1-BCC

8. Sage with preset charges on water
* Water gets preset charges
* Ions get library charges
* Non-water solvent gets AM1-BCC
* Ligands get AM1-BCC

9. Sage with (ligand) virtual site parameters
* Water gets preset charges
* Ions get library charges
* Non-water solvent gets AM1-BCC
* Ligand heavy atoms get AM1-BCC and charge increments
* Virtual sites get charge increments

10. AM1-with-custom-BCCs Sage with ligand and ions water
* Water gets library charges
* Ions get library charges
* Ligand gets charge increments

11. Sage with NAGL and ligand in vacuum
* Ligand gets NAGL charges

12. Sage with NAGL and water molecules (library precedence)
* Water gets library charges (precedence over NAGL)

13. Sage with NAGL mixed topology (ligand and water)
* Ligand gets NAGL charges
* Water gets library charges

14. Sage with NAGL and preset charges on ligand
* Ligand gets preset charges (precedence over NAGL)

Other details
* Specifics of charge method (AM1-BCC, AM1-BCC-ELF10, AM1-BCC via NAGL)
* Molecules with preset charges can be similar but not exact enough
* Preset charges and virtual sites is underspecified and cannot be tested
"""

# TODO: Revisit test cases given that NAGL charges are default (2.3.0+, 3.0.0, etc.)
AM1BCC_KEY = "am1bccelf10" if OPENEYE_AVAILABLE else "am1bcc"
NAGL_KEY = "openff-gnn-am1bcc-1.0.0.pt"

INCHI_KEYS = {
    "ethanol": "LFQSCWFLJHTTHZ-UHFFFAOYSA-N",
    "water": "XLYOFNOQVPJJNP-UHFFFAOYSA-N",
    "sodium": "FKNQFGJONOIPTF-UHFFFAOYSA-N",
    "chloride": "VEXZGXHMUGYJMC-UHFFFAOYSA-M",
    "methylamine": "BAVYZALUXZFZLV-UHFFFAOYSA-N",
    "protein": "QIVKPXNXCAPCCO-WDSKDSINSA-N",
}


def map_methods_to_inchikey(caplog: pytest.LogCaptureFixture) -> dict[str, list[int]]:
    """
    Map partial charge assignment methods to (sorted) molecule InChI.
    """
    debug = defaultdict(list)

    for record in caplog.records:
        if record.name.startswith("openff.interchange"):
            # This is a specific one-off logging event not relevant to this function
            if "Could not generate InChIKey for molecule with Hill formula" in record.msg:
                continue
            assert record.levelname == "DEBUG", "Only DEBUG logs are expected."
        else:
            # skip logged warnings from upstreams/other packages
            continue

        message = record.msg

        if message.startswith("Charge section LibraryCharges"):
            _, molecule = message.split(", ")

            debug["library"].append(molecule.split("InChIKey ")[-1])

        elif message.startswith("Charge section ToolkitAM1BCC"):
            _, method, molecule = message.split(", ")

            assert method.split()[-1] == AM1BCC_KEY, f"Expected method {AM1BCC_KEY} but got {method.split()[-1]}"

            debug[method.split()[-1]].append(molecule.split("InChIKey ")[-1])

        elif message.startswith("Charge section NAGLCharges"):
            _, _, molecule = message.split(", ")

            debug["NAGL"].append(molecule.split("InChIKey ")[-1])

        # without also pulling the virtual site - particle mapping (which is different for each engine)
        # it's hard to store more debugrmation than the orientation atoms that are affected by each
        # virtual site's charges
        elif message.startswith("Charge section VirtualSites"):
            orientation_atoms: list[int] = [
                int(value.strip()) for value in re.findall(r"\((.*?)\)", message)[0].split(",")
            ]

            for atom in orientation_atoms:
                debug["orientation"].append(atom)

        elif message.startswith("Preset charges"):
            debug["preset"].append(message.split("InChIKey ")[-1])

        elif message.startswith("Charge section ChargeIncrementModel"):
            if "using charge method" in message:
                _, method, molecule = message.split(", ")
                debug[f"chargeincrements_{method.split()[-1]}"].append(
                    molecule.split("InChIKey ")[-1],
                )

            elif "applying charge increment" in message:
                # TODO: Store the "other" atoms?
                debug["chargeincrements"].append(int(message.split("atom ")[1].split(" ")[0]))

        else:
            raise ValueError(f"Unexpected log message {message}")

    return {key: sorted(val) for key, val in debug.items()}


def _ensure_pre_nagl_sage(sage: ForceField) -> ForceField:
    """Revert a verison of Sage to pre-NAGL by swapping out NAGLChharges handler for ToolkitAM1BCC."""
    try:
        sage.deregister_parameter_handler("NAGLCharges")
        sage.get_parameter_handler("ToolkitAM1BCC")

        return sage

    except KeyError:
        return sage


@pytest.fixture
def sage_with_nagl_chargeincrements(sage_no_nagl):
    from openff.toolkit.typing.engines.smirnoff.parameters import ChargeIncrementModelHandler, ChargeIncrementType

    sage_no_nagl.register_parameter_handler(
        parameter_handler=ChargeIncrementModelHandler(
            version=0.3,
            partial_charge_method=NAGL_KEY,
        ),
    )

    # Add dummy "BCCs" for testing, even though this model has BCCs built into
    # the partial charge assignment method
    sage_no_nagl["ChargeIncrementModel"].add_parameter(
        parameter=ChargeIncrementType(
            smirks="[#6:1]-[#1:2]",
            charge_increment=[
                "0.1 elementary_charge",
                "-0.1 elementary_charge",
            ],
        ),
    )

    return sage_no_nagl


@pytest.fixture
def ions() -> Topology:
    return Topology.from_molecules(
        [
            Molecule.from_smiles(smiles)
            for smiles in [
                "[Na+]",
                "[Cl-]",
                "[Na+]",
                "[Cl-]",
            ]
        ],
    )


@pytest.fixture
def ligand() -> Molecule:
    return Molecule.from_mapped_smiles("[C:1]([C:2]([O:3][H:9])([H:7])[H:8])([H:4])([H:5])[H:6]")


@pytest.fixture
def solvent() -> Molecule:
    return Molecule.from_mapped_smiles("[H:3][C:1]([H:4])([H:5])[N:2]([H:6])[H:7]")


def ligand_in_solvent(ligand, solvent) -> Topology:
    return Topology.from_molecules(
        [
            ligand,
            solvent,
            solvent,
            solvent,
        ],
    )


@pytest.fixture
def water_and_ions(water, ions) -> Topology:
    topology = Topology.from_molecules(3 * [water])
    for molecule in ions.molecules:
        topology.add_molecule(molecule)

    return topology


@pytest.fixture
def ligand_and_water_and_ions(ligand, water_and_ions) -> Topology:
    topology = ligand.to_topology()

    for molecule in water_and_ions.molecules:
        topology.add_molecule(molecule)

    return topology


"""
0.xSage with just a ligand
1.xSage with a ligand in water/ions
2.xSage with a ligand in non-water solvent
3.xSage with a ligand in mixed solvent
4.xff14SB with Sage
5.xff14SB with Sage and preset charges on Protein A
6.xSage with ligand and OPC water
7.xSage with preset charges on ligand A
8.xSage with preset charges on water
9.xSage with (ligand) virtual site parameters
10.xAM1-with-custom-BCCs Sage with ligand and ions water
11.xSage with NAGL and ligand in vacuum
12.xSage with NAGL and water molecules (library precedence)
13.xSage with NAGL mixed topology (ligand and water)
14.xSage with NAGL and preset charges on ligand
"""


def test_case0(caplog, sage_no_nagl, ligand):
    with caplog.at_level(logging.DEBUG, logger="openff.interchange"):
        sage_no_nagl.create_interchange(ligand.to_topology())

        debug = map_methods_to_inchikey(caplog)

        # atoms 0 through 8 are ethanol, getting AM1-BCC
        assert debug[AM1BCC_KEY] == [INCHI_KEYS["ethanol"]]


def test_case1(caplog, sage_no_nagl, ligand_and_water_and_ions):
    with caplog.at_level(logging.DEBUG, logger="openff.interchange"):
        sage_no_nagl.create_interchange(ligand_and_water_and_ions)

        debug = map_methods_to_inchikey(caplog)

        # atoms 0 through 8 are ethanol, getting AM1-BCC
        assert debug[AM1BCC_KEY] == [INCHI_KEYS["ethanol"]]

        # atoms 9 through 21 are water + ions, getting library charges
        assert debug["library"] == [
            INCHI_KEYS["sodium"],
            INCHI_KEYS["chloride"],
            INCHI_KEYS["water"],
        ]


def test_case2(caplog, sage_no_nagl, ligand, solvent):
    topology = Topology.from_molecules([ligand, solvent, solvent, solvent])

    with caplog.at_level(logging.DEBUG, logger="openff.interchange"):
        sage_no_nagl.create_interchange(topology)

        debug = map_methods_to_inchikey(caplog)

        # everything should get AM1-BCC charges
        assert debug[AM1BCC_KEY] == [INCHI_KEYS["methylamine"], INCHI_KEYS["ethanol"]]


def test_case3(caplog, sage_no_nagl, ligand_and_water_and_ions, solvent):
    for index in range(3):
        ligand_and_water_and_ions.add_molecule(solvent)

    ligand_and_water_and_ions.molecule(0).assign_partial_charges(partial_charge_method="gasteiger")

    with caplog.at_level(logging.DEBUG, logger="openff.interchange"):
        sage_no_nagl.create_interchange(
            ligand_and_water_and_ions,
        )

        debug = map_methods_to_inchikey(caplog)

        # atoms 0 through 8 are ethanol, getting AM1-BCC,
        # and also solvent molecules (starting at index 22)
        assert debug[AM1BCC_KEY] == [INCHI_KEYS["methylamine"], INCHI_KEYS["ethanol"]]

        # atoms 9 through 21 are water + ions, getting library charges
        assert debug["library"] == [
            INCHI_KEYS["sodium"],
            INCHI_KEYS["chloride"],
            INCHI_KEYS["water"],
        ]


@pytest.mark.slow
@pytest.mark.parametrize("preset_on_protein", [True, False])
def test_cases4_5(caplog, ligand_and_water_and_ions, preset_on_protein):
    ff = ForceField("ff14sb_off_impropers_0.0.3.offxml", "openff-2.0.0.offxml")

    complex = get_protein("MainChain_ALA_ALA").to_topology()

    for molecule in ligand_and_water_and_ions.molecules:
        complex.add_molecule(molecule)

    if preset_on_protein:
        complex.molecule(0).assign_partial_charges(partial_charge_method="zeros")

    with caplog.at_level(logging.DEBUG, logger="openff.interchange"):
        if preset_on_protein:
            ff.create_interchange(complex, charge_from_molecules=[complex.molecule(0)])
        else:
            ff.create_interchange(complex)

        debug = map_methods_to_inchikey(caplog)

        assert debug[AM1BCC_KEY] == [INCHI_KEYS["ethanol"]]

        if preset_on_protein:
            # protein gets preset charges
            assert debug["preset"] == [
                INCHI_KEYS["protein"],
            ]

            # everything after the protein and ligand should get library charges
            assert debug["library"] == [
                INCHI_KEYS["sodium"],
                INCHI_KEYS["chloride"],
                INCHI_KEYS["water"],
            ]
        else:
            # the protein and everything after the ligand should get library charges
            assert debug["library"] == [
                INCHI_KEYS["sodium"],
                INCHI_KEYS["protein"],
                INCHI_KEYS["chloride"],
                INCHI_KEYS["water"],
            ]


def test_case6(caplog, ligand, water):
    force_field = ForceField("openff-2.0.0.offxml", "opc.offxml")

    topology = Topology.from_molecules([ligand, water, water, water])

    with caplog.at_level(logging.DEBUG, logger="openff.interchange"):
        force_field.create_interchange(topology)

        debug = map_methods_to_inchikey(caplog)

        # atoms 0 through 8 are ethanol, getting AM1-BCC charges
        assert debug[AM1BCC_KEY] == [INCHI_KEYS["ethanol"]]

        # atoms 9 through 17 are water atoms, getting library charges
        assert debug["library"] == [INCHI_KEYS["water"]]

        # particles 18 through 20 are water virtual sites, but the current logging strategy
        # makes it hard to match these up (and the particle indices are different OpenMM/GROMACS/etc)

        # can still check that orientation atoms are subject to virtual site
        # charge increments (even if the increment is +0.0 e)
        assert debug["orientation"] == [*range(9, 18)]


def test_case7(caplog, sage_no_nagl, ligand_and_water_and_ions):
    ligand_and_water_and_ions.molecule(0).assign_partial_charges(partial_charge_method="gasteiger")

    with caplog.at_level(logging.DEBUG, logger="openff.interchange"):
        sage_no_nagl.create_interchange(
            ligand_and_water_and_ions,
            charge_from_molecules=[ligand_and_water_and_ions.molecule(0)],
        )

        debug = map_methods_to_inchikey(caplog)

        # atoms 0 through 8 are ethanol, getting preset charges
        assert debug["preset"] == [INCHI_KEYS["ethanol"]]

        # atoms 9 through 21 are water + ions, getting library charges
        assert debug["library"] == [
            INCHI_KEYS["sodium"],
            INCHI_KEYS["chloride"],
            INCHI_KEYS["water"],
        ]


def test_case8(caplog, sage_no_nagl, water_and_ions):
    water_and_ions.molecule(0).assign_partial_charges(partial_charge_method="gasteiger")

    with caplog.at_level(logging.DEBUG, logger="openff.interchange"):
        sage_no_nagl.create_interchange(
            water_and_ions,
            charge_from_molecules=[water_and_ions.molecule(0)],
        )

        debug = map_methods_to_inchikey(caplog)

        # atoms 0 through 8 are water, getting preset charges
        assert debug["preset"] == [INCHI_KEYS["water"]]

        # atoms 9 through 12 are ions, getting library charges
        assert debug["library"] == [INCHI_KEYS["sodium"], INCHI_KEYS["chloride"]]


def test_case9(caplog, sage_with_bond_charge):
    with caplog.at_level(logging.DEBUG, logger="openff.interchange"):
        _ensure_pre_nagl_sage(sage_with_bond_charge).create_interchange(
            Molecule.from_mapped_smiles(
                "[H:3][C:1]([H:4])([H:5])[Cl:2]",
            ).to_topology(),
        )

        debug = map_methods_to_inchikey(caplog)

        # atoms 0 through 5 are ligand, getting AM1-BCC charges
        assert debug[AM1BCC_KEY] == ["NEHMKBQYUWJMIP-UHFFFAOYSA-N"]

        # atoms 0 and 1 are the orientation atoms of the sigma hole virtual site
        assert debug["orientation"] == [0, 1]


def test_case10(caplog, sage_with_nagl_chargeincrements, ligand):
    pytest.importorskip("openff.nagl")
    pytest.importorskip("rdkit")

    with caplog.at_level(logging.DEBUG, logger="openff.interchange"):
        with toolkit_registry_manager(
            toolkit_registry=ToolkitRegistry(
                toolkit_precedence=[NAGLToolkitWrapper, RDKitToolkitWrapper],
            ),
        ):
            _ensure_pre_nagl_sage(sage_with_nagl_chargeincrements).create_interchange(
                ligand.to_topology(),
            )

        debug = map_methods_to_inchikey(caplog)

        # atoms 0 through 8 are ligand, getting NAGL charges
        assert debug[f"chargeincrements_{NAGL_KEY}"] == ["LFQSCWFLJHTTHZ-UHFFFAOYSA-N"]

        # TODO: These are logged symmetrically (i.e. hydrogens are listed)
        # even though the charges appear to be correct, assert should
        # simply by == [0, 1] since the hydrogens shouldn't be caught
        assert 0 in debug["chargeincrements"]
        assert 1 in debug["chargeincrements"]

        # the standard AM1-BCC should not have ran
        assert AM1BCC_KEY not in debug


def test_case11(caplog, sage, ligand):
    """Test that NAGL charge assignment is properly logged."""
    pytest.importorskip("openff.nagl")

    with caplog.at_level(logging.DEBUG, logger="openff.interchange"):
        sage.create_interchange(ligand.to_topology())

        debug = map_methods_to_inchikey(caplog)

        # Should log NAGL charges for all atoms
        assert "NAGL" in debug
        assert debug["NAGL"] == [INCHI_KEYS["ethanol"]]


def test_case12(caplog, sage, water):
    """Test logging when LibraryCharges takes precedence over NAGL."""
    pytest.importorskip("openff.nagl")

    topology = Topology.from_molecules([water, water])

    with caplog.at_level(logging.DEBUG, logger="openff.interchange"):
        sage.create_interchange(topology)

        debug = map_methods_to_inchikey(caplog)

        # Water should get library charges, not NAGL
        assert debug["library"] == [INCHI_KEYS["water"]]  # 2 water molecules
        assert "NAGL" not in debug


def test_case13(caplog, sage, ligand, water):
    """Test logging with mixed molecule types (some library, some NAGL)."""
    pytest.importorskip("openff.nagl")

    topology = Topology.from_molecules([ligand, water])

    with caplog.at_level(logging.DEBUG, logger="openff.interchange"):
        sage.create_interchange(topology)

        debug = map_methods_to_inchikey(caplog)

        # Ligand should get NAGL charges
        assert debug["NAGL"] == [INCHI_KEYS["ethanol"]]

        # Water should get library charges
        assert debug["library"] == [INCHI_KEYS["water"]]


def test_case14(caplog, sage, ligand):
    """Test logging when preset charges are used instead of NAGL."""
    pytest.importorskip("openff.nagl")

    ligand.assign_partial_charges("gasteiger")

    with caplog.at_level(logging.DEBUG, logger="openff.interchange"):
        sage.create_interchange(
            ligand.to_topology(),
            charge_from_molecules=[ligand],
        )

        debug = map_methods_to_inchikey(caplog)

        # Should use preset charges, not NAGL
        assert debug["preset"] == [INCHI_KEYS["ethanol"]]
        assert "NAGL" not in debug


def test_inchi_fallback(caplog, sage):
    """Test that molecules that fail InChIKey generation are still logged in some way."""
    from openff.toolkit.utils.toolkits import OpenEyeToolkitWrapper

    # TODO: Might be a toolkit bug that needs to be worked around
    with caplog.at_level(logging.DEBUG, logger="openff.interchange"):
        sage.create_interchange(
            Molecule.from_smiles(342 * "C").to_topology(),
        )

        debug = map_methods_to_inchikey(caplog)

        if OpenEyeToolkitWrapper.is_available():
            assert debug["NAGL"] == ["UNKNOWN_INCHIKEY"]
        else:
            # but RDKit can generate an InChIKey
            assert debug["NAGL"] == ["FGIJOQKVOYCTPJ-UHFFFAOYSA-N"]



import json

file_name = "regression_tests/differences.json"
input_path = "regression_tests/small-molecule/input-topologies.json"

differences = dict()

try:
    with open(file_name) as file:
        differences.update(json.loads(file.read()))

except FileNotFoundError:
    print(f"File {file_name} not found.")

if len(differences) > 0:
    with open(input_path) as file:
        inputs = json.loads(file.read())

    for molecule, difference in differences.items():
        print(inputs[int(molecule)])
        print("\t" + str(difference))

    exit(1)



import pytest
from openff.toolkit import Quantity
from openff.toolkit.topology import ValenceDict
from openff.toolkit.typing.engines.smirnoff.parameters import ConstraintHandler
from openff.toolkit.utils.exceptions import SMIRNOFFSpecError

from openff.interchange.exceptions import MissingParametersError


@pytest.fixture
def distanceless_bond_constraints():
    handler = ConstraintHandler(version=0.3)
    handler.add_parameter(
        {
            "smirks": "[*:1]~[*:2]",  # all bonds!
            "id": "distanceless_bond",
        },
    )

    return handler


@pytest.fixture
def bond_constraints_with_distance():
    handler = ConstraintHandler(version=0.3)
    handler.add_parameter(
        {
            "smirks": "[*:1]~[*:2]",
            "id": "bond_with_distance",
            "distance": Quantity("0.123456789 nm"),
        },
    )

    return handler


@pytest.fixture
def distanceless_angle_constraints():
    handler = ConstraintHandler(version=0.3)
    handler.add_parameter(
        {
            "smirks": "[*:1]~[*]~[*:2]",  # all angles!
            "id": "distanceless_angle",
        },
    )

    return handler


@pytest.fixture
def angle_constraints_with_distance():
    handler = ConstraintHandler(version=0.3)
    handler.add_parameter(
        {
            "smirks": "[*:1]~[*]~[*:2]",  # all angles!
            "id": "angle_with_distance",
            "distance": Quantity("0.22222222 nm"),
        },
    )

    return handler


class TestConstraints:
    def test_with_bonds_with_distanceless_constraints(
        self,
        sage,
        distanceless_bond_constraints,
        ethanol,
    ):
        """Bonds specified, constraints without distance, length taken from bonds."""
        openmm = pytest.importorskip("openmm")

        sage.deregister_parameter_handler("Constraints")
        sage.register_parameter_handler(distanceless_bond_constraints)

        # add in bond parameters which are also constrained, since it's easier and a little safer
        # than looking up what they *should* be
        system = sage.create_interchange(ethanol.to_topology()).to_openmm_system(add_constrained_forces=True)

        assert system.getNumConstraints() == ethanol.n_bonds

        constraint_distances = ValenceDict()
        bond_distances = ValenceDict()

        for constraint_index in range(system.getNumConstraints()):
            atom1_index, atom2_index, distance = system.getConstraintParameters(constraint_index)

            assert "0.123456789" not in str(distance)

            constraint_distances[(atom1_index, atom2_index)] = distance

        bond_force = next(force for force in system.getForces() if isinstance(force, openmm.HarmonicBondForce))

        for bond_index in range(bond_force.getNumBonds()):
            atom1_index, atom2_index, length, _ = bond_force.getBondParameters(bond_index)

            bond_distances[(atom1_index, atom2_index)] = length

        # some bonds (C-C, C-O) don't have constraints, so can't compare entire dicts
        # but all constraints should have corresponding bond parameters
        for constrained_pair, constrained_distance in constraint_distances.items():
            assert constrained_distance == bond_distances[constrained_pair]

    def test_constraint_distances_override_bond_distances(
        self,
        sage,
        bond_constraints_with_distance,
        ethanol,
    ):
        """Bonds specified, constraints without distance, length taken from bonds."""
        openmm = pytest.importorskip("openmm")

        # replace original constraints with a single wildcard constraint, since it's
        # easier and a little safer than looking up what each distance  *should* be

        sage.deregister_parameter_handler("Constraints")
        sage.register_parameter_handler(bond_constraints_with_distance)
        system = sage.create_interchange(ethanol.to_topology()).to_openmm_system(add_constrained_forces=True)

        assert system.getNumConstraints() == ethanol.n_bonds

        for constraint_index in range(system.getNumConstraints()):
            _, _, distance = system.getConstraintParameters(constraint_index)

            assert "0.12345678" in str(distance)

        for force in system.getForces():
            if type(force) is not openmm.HarmonicBondForce:
                continue
            for bond_index in range(force.getNumBonds()):
                _, _, length, _ = force.getBondParameters(bond_index)

                assert "0.12345678" not in str(length)

    def test_distanceless_constraints_without_bonds_error(
        self,
        sage,
        distanceless_bond_constraints,
        ethanol,
    ):
        """When constraints are speicifed without distances, but no bonds are present, error."""
        sage.deregister_parameter_handler("Constraints")
        sage.register_parameter_handler(distanceless_bond_constraints)
        sage.deregister_parameter_handler("Bonds")

        # this test short-circuits before anything OpenMM is called, so it could live elsewhere
        with pytest.raises(
            MissingParametersError,
            match=r"The distance of this constraint is not specified.",
        ):
            sage.create_openmm_system(ethanol.to_topology())

    def test_constraints_with_distances_without_bonds(
        self,
        sage,
        bond_constraints_with_distance,
        ethanol,
    ):
        """When constraints are speicifed with distances, but no bonds are present, still sets constraints."""
        openmm = pytest.importorskip("openmm")

        sage.deregister_parameter_handler("Constraints")
        sage.register_parameter_handler(bond_constraints_with_distance)
        sage.deregister_parameter_handler("Bonds")

        system = sage.create_interchange(ethanol.to_topology()).to_openmm_system()

        assert system.getNumConstraints() == ethanol.n_bonds

        for constraint_index in range(system.getNumConstraints()):
            _, _, distance = system.getConstraintParameters(constraint_index)

            assert "0.12345678" in str(distance)

        for force in system.getForces():
            if type(force) is openmm.HarmonicBondForce:
                raise Exception("there should not be a bond force")

    @pytest.mark.parametrize("remove_angles", [True, False])
    def test_angles_with_distanceless_constraints(
        self,
        distanceless_angle_constraints,
        sage,
        ethanol,
        remove_angles,
    ):
        """
        Test that angle-like constraints without specified distance raise an error,
        whether or not there are angle parameters.
        """
        sage.deregister_parameter_handler("Constraints")
        sage.register_parameter_handler(distanceless_angle_constraints)

        if remove_angles:
            sage.deregister_parameter_handler("Angles")

        with pytest.raises(
            SMIRNOFFSpecError,
            match=r"0, 2.*unsupported in the SMIRNOFF specification",
        ):
            sage.create_openmm_system(ethanol.to_topology())

    def test_constraint_distances_override_angle_geometry(
        self,
        sage,
        ethanol,
        angle_constraints_with_distance,
    ):
        sage.deregister_parameter_handler("Constraints")
        sage.register_parameter_handler(angle_constraints_with_distance)

        system = sage.create_interchange(ethanol.to_topology()).to_openmm_system(add_constrained_forces=True)

        assert system.getNumConstraints() == ethanol.n_angles

        for constraint_index in range(system.getNumConstraints()):
            _, _, distance = system.getConstraintParameters(constraint_index)

            assert "0.222222" in str(distance)

        # angle parameters don't store 1-3 distance, so can't cleanly compare geometries vs. constraint distance

    def test_constraints_with_distances_without_angles(
        self,
        sage,
        ethanol,
        angle_constraints_with_distance,
    ):
        """When constraints are speicifed with distances, but no angles are present, still sets constraints."""
        openmm = pytest.importorskip("openmm")

        sage.deregister_parameter_handler("Constraints")
        sage.register_parameter_handler(angle_constraints_with_distance)
        sage.deregister_parameter_handler("Angles")

        system = sage.create_interchange(ethanol.to_topology()).to_openmm_system()

        assert system.getNumConstraints() == ethanol.n_angles

        for constraint_index in range(system.getNumConstraints()):
            _, _, distance = system.getConstraintParameters(constraint_index)

            assert "0.222222" in str(distance)

        for force in system.getForces():
            if type(force) is openmm.HarmonicAngleForce:
                raise Exception("there should not be an angle force")



"""Utilities for processing and interfacing with mBuild models."""

from typing import TYPE_CHECKING

from openff.toolkit import Molecule, Topology
from openff.utilities.utilities import has_package, requires_package

if has_package("mbuild") or TYPE_CHECKING:
    import mbuild


@requires_package("mbuild")
def offmol_to_compound(off_mol: "Molecule") -> "mbuild.Compound":
    """
    Convert an OpenFF Molecule into an mBuild Compound.

    Examples
    --------
    .. code-block:: python

        >>> from openff.toolkit import Molecule
        >>> from openff.interchange.components.mbuild import offmol_to_compound
        >>> mol = Molecule.from_smiles("CCO")
        >>> compound = offmol_to_compound(mol)
        >>> type(compound), compound.n_particles, compound.n_bonds
        (<class 'mbuild.compound.Compound'>, 9, 8)

    """
    if not off_mol.has_unique_atom_names:
        off_mol.generate_unique_atom_names()

    if off_mol.n_conformers == 0:
        off_mol.generate_conformers(n_conformers=1)

    comp = mbuild.Compound()
    comp.name = off_mol.name

    for a in off_mol.atoms:
        atom_comp = mbuild.Particle(name=a.symbol)
        comp.add(atom_comp, label=a.name)

    for b in off_mol.bonds:
        comp.add_bond((comp[b.atom1_index], comp[b.atom2_index]))

    comp.xyz = off_mol.conformers[0].m_as("nanometer")

    return comp


@requires_package("mbuild")
def offtop_to_compound(off_top: "Topology") -> "mbuild.Compound":
    """
    Convert an OpenFF Topology into an mBuild Compound.

    Examples
    --------
    .. code-block:: python

        >>> from openff.toolkit import Molecule, Topology
        >>> from openff.interchange.components.mbuild import offtop_to_compound
        >>> ethanol = Molecule.from_smiles("CCO")
        >>> ethanol.name = "ETH"
        >>> methane = Molecule.from_smiles("C")
        >>> methane.name = "MET"
        >>> top = Topology.from_molecules([ethanol, ethanol, methane, methane])
        >>> compound = offtop_to_compound(top)
        >>> type(compound), len(compound.children), compound.n_particles, compound.n_bonds
        (<class 'mbuild.compound.Compound'>, 4, 28, 24)

    """
    return mbuild.Compound(
        subcompounds=[offmol_to_compound(molecule) for molecule in off_top.molecules],
    )



"""Custom warnings used in Interchange."""


class InterchangeDeprecationWarning(UserWarning):
    """Warning for deprecated portions of the Interchange API."""


class SwitchingFunctionNotImplementedWarning(UserWarning):
    """Exporting to an engine that does not implement a switching function."""


class MissingPositionsWarning(UserWarning):
    """Warning for when positions are likely needed but missing."""


class NonbondedSettingsWarning(UserWarning):
    """Warning for when details of nonbonded implementations get messy."""


class ForceFieldModificationWarning(UserWarning):
    """Warning for when a ForceField is modified."""


class PresetChargesAndVirtualSitesWarning(UserWarning):
    """Warning when possibly using preset charges and virtual sites together."""


class InterchangeCombinationWarning(UserWarning):
    """Warning for when combining Interchange objects."""



import pytest
from openff.toolkit import Quantity

from openff.interchange._tests import MoleculeWithConformer, needs_lmp
from openff.interchange.constants import kj_mol
from openff.interchange.drivers.lammps import _process, get_lammps_energies


class TestProcess:
    @pytest.fixture
    def dummy_energies(self):
        energies = [Quantity(val, kj_mol) for val in [2.0, 1.0, 1.5, 0.5, -100.0, 0.1, -4, -400]]

        return {
            "Bond": energies[0],
            "Angle": energies[1],
            "ProperTorsion": energies[2],
            "ImproperTorsion": energies[3],
            "vdW": energies[4],
            "DispersionCorrection": energies[5],
            "ElectrostaticsShort": energies[6],
            "ElectrostaticsLong": energies[7],
        }

    def test_simple(self, dummy_energies):
        processed = _process(dummy_energies)

        for key in ["Bond", "Angle"]:
            assert processed[key] == dummy_energies[key]

        assert processed["Torsion"] == (dummy_energies["ProperTorsion"] + dummy_energies["ImproperTorsion"])
        assert processed["vdW"] == (dummy_energies["vdW"] + dummy_energies["DispersionCorrection"])
        assert processed["Electrostatics"] == (
            dummy_energies["ElectrostaticsShort"] + dummy_energies["ElectrostaticsLong"]
        )

    @pytest.mark.skip("Not implemented yet")
    def test_detailed(self, dummy_energies):
        assert "ImproperTorsion" in _process(
            dummy_energies,
            detailed=True,
        )


class TestLAMMPSDriver:
    @needs_lmp
    def test_can_write_without_torsions(self, sage):
        topology = MoleculeWithConformer.from_smiles("N").to_topology()
        topology.box_vectors = Quantity([4, 4, 4], "nanometer")

        assert get_lammps_energies(sage.create_interchange(topology))["Bond"].m == 0.0



"""Import guard."""

# openff/interchange/foyer.py causes "import foyer" to not crash, which breaks has_packge("foyer")
# and breaks @skip_if_missing("foyer"), so need a more specific way of checking Foyer installation
try:
    from foyer import Forcefield  # noqa: F401

    has_foyer = True
except (ImportError, AttributeError):
    has_foyer = False



from openff.toolkit import Topology

from openff.interchange._tests import MoleculeWithConformer


def compare_topologies(topology1, topology2):
    """Compare two topologies for equality."""
    assert topology1.to_dict() == topology2.to_dict()


class TestPDBRoundtripVsToolkit:
    def test_roundtrip_vs_toolkit(self, sage, tmp_path):
        """Test that Interchange.to_pdb can roundtrip with Topology.from_pdb."""
        h2po4 = MoleculeWithConformer.from_smiles("[O-]P(=O)(O)O")

        out = sage.create_interchange(h2po4.to_topology())

        # include_virtual_sites=True errors since there are no virtual sites
        out.to_pdb("interchange.pdb", include_virtual_sites=False)
        h2po4.to_topology().to_file("toolkit.pdb")

        compare_topologies(
            Topology.from_pdb("interchange.pdb", unique_molecules=[h2po4]),
            Topology.from_pdb("toolkit.pdb", unique_molecules=[h2po4]),
        )



import json

from openff.interchange.common._nonbonded import ElectrostaticsCollection


def test_properties_on_child_collections_serialized():
    blob = ElectrostaticsCollection(scale_14=2.1).model_dump_json()

    assert json.loads(blob)["scale_14"] == 2.1

    assert ElectrostaticsCollection.model_validate_json(blob).scale_14 == 2.1
